"""Creation of 'corrgram' correlation matrix plots"""


from __future__ import annotations

import json
import logging
import math
import os

import libtbx.phil

logger = logging.getLogger(__name__)

# PHIL options for corrgrams
phil_str = """
    correlation_plot
      .expert_level = 1
    {
      filename = None
        .type = str
        .help = "The base filename for output of plots of parameter"
                "correlations. A file extension may be added to control"
                "the type of output file, if it is one of matplotlib's"
                "supported types. A JSON file with the same base filename"
                "will also be created, containing the correlation matrix and"
                "column labels for later inspection, replotting etc."

      col_select = None
        .type = strings
        .help = "Specific columns to include in the plots of parameter"
                "correlations, either specified by parameter name or 0-based"
                "column index. Defaults to all columns."
                "This option is useful when there is a large number of"
                "parameters"

      steps = None
        .type = ints(value_min=0)
        .help = "Steps for which to make correlation plots. By default only"
                "the final step is plotted. Uses zero-based numbering, so"
                "the first step is numbered 0."
    }
"""

phil_scope = libtbx.phil.parse(phil_str)


def corrgram(corrmat, labels):
    """Create a correlation matrix plot or 'corrgram' for the provided
    correlation matrix and row/column labels. Inspired by R's corrplot and
    https://github.com/louridas/corrplot/blob/master/corrplot.py"""

    try:  # is corrmat a scitbx matrix?
        corrmat = corrmat.as_flex_double_matrix()
    except AttributeError:  # assume it is already a flex double matrix
        pass
    assert corrmat.is_square_matrix()

    nr = corrmat.all()[0]
    assert nr == len(labels)

    try:
        import matplotlib

        matplotlib.use("Agg")
        import matplotlib.cm as cm
        import matplotlib.pyplot as plt
    except ImportError as e:
        logger.info("matplotlib modules not available " + str(e), exc_info=True)
        return None

    plt.figure(1)
    ax = plt.subplot(1, 1, 1, aspect="equal")
    clrmap = cm.get_cmap("bwr")

    for x in range(nr):
        for y in range(nr):
            d = corrmat[x, y]
            d_abs = abs(d)
            circ = plt.Circle((x, y), radius=0.9 * math.sqrt(d_abs) / 2)
            circ.set_edgecolor("white")
            # put data into range [0,1] and invert so that 1 == blue and 0 == red
            facecolor = 1 - (0.5 * d + 0.5)
            circ.set_facecolor(clrmap(facecolor))
            ax.add_artist(circ)
    ax.set_xlim(-0.5, nr - 0.5)
    ax.set_ylim(-0.5, nr - 0.5)

    ax.xaxis.tick_top()
    xtickslocs = list(range(len(labels)))
    ax.set_xticks(xtickslocs)
    ax.set_xticklabels(labels, rotation=30, fontsize="small", ha="left")

    ax.invert_yaxis()
    ytickslocs = list(range(len(labels)))
    ax.set_yticks(ytickslocs)
    ax.set_yticklabels(labels, fontsize="small")

    xtickslocs = [loc + 0.5 for loc in xtickslocs]
    ax.set_xticks(xtickslocs, minor=True)
    ytickslocs = [loc + 0.5 for loc in ytickslocs]
    ax.set_yticks(ytickslocs, minor=True)
    plt.grid(color="0.8", which="minor", linestyle="-")

    # suppress major tick marks
    ax.tick_params(which="major", width=0)

    # need this otherwise text gets clipped
    plt.tight_layout()

    return plt


def create_correlation_plots(refiner, params):
    root, ext = os.path.splitext(params.correlation_plot.filename)
    if not ext:
        ext = ".pdf"

    history = refiner.history
    steps = params.correlation_plot.steps
    if steps is None:
        steps = [history.get_nrows() - 1]

    # extract individual column names or indices
    col_select = params.correlation_plot.col_select

    num_plots = 0
    for step in steps:
        fname_base = root
        if len(steps) > 1:
            fname_base += "_step%02d" % step

        corrmats, labels = refiner.get_parameter_correlation_matrix(step, col_select)
        if [corrmats, labels].count(None) == 0:

            for resid_name, corrmat in corrmats.items():
                plot_fname = fname_base + "_" + resid_name + ext
                plt = corrgram(corrmat, labels)
                if plt is not None:
                    logger.info(f"Saving parameter correlation plot to {plot_fname}")
                    plt.savefig(plot_fname)
                    plt.close()
                    num_plots += 1
            mat_fname = fname_base + ".json"
            with open(mat_fname, "w") as handle:
                for k, corrmat in corrmats.items():
                    corrmats[k] = corrmat.as_scitbx_matrix().as_list_of_lists()
                logger.info(f"Saving parameter correlation matrices to {mat_fname}")
                json.dump({"corrmats": corrmats, "labels": labels}, handle)

    if num_plots == 0:
        msg = (
            "Sorry, no parameter correlation plots were produced. Please set "
            "track_parameter_correlation=True to ensure correlations are "
            "tracked, and make sure correlation_plot.col_select is valid."
        )
        logger.info(msg)



from __future__ import annotations

import math
import pickle
import random

from scitbx import matrix

from dials.model.data import Shoebox


def random_shoeboxes(num, mask=False):
    for i in range(num):
        x0 = random.randint(0, 100)
        y0 = random.randint(0, 100)
        z0 = random.randint(0, 100)
        x1 = random.randint(x0 + 5, x0 + 20)
        y1 = random.randint(y0 + 5, y0 + 20)
        z1 = random.randint(z0 + 5, z0 + 20)
        bbox = (x0, x1, y0, y1, z0, z1)
        xc0 = (x1 + x0) / 2.0
        yc0 = (y1 + y0) / 2.0
        zc0 = (z1 + z0) / 2.0
        xc = random.uniform(xc0 - 1, xc0 + 1)
        yc = random.uniform(yc0 - 1, yc0 + 1)
        zc = random.uniform(zc0 - 1, zc0 + 1)
        centre = (xc, yc, zc)
        intensity = random.randint(10, 10000)
        shoebox = generate_shoebox(bbox, centre, intensity, mask=mask)
        yield (shoebox, (centre, intensity))


def generate_shoebox(bbox, centre, intensity, mask=False):
    from dials.algorithms.shoebox import MaskCode

    shoebox = Shoebox()
    shoebox.bbox = bbox
    shoebox.allocate()
    for i in range(len(shoebox.mask)):
        shoebox.mask[i] = MaskCode.Valid | MaskCode.Foreground
    shoebox.data = gaussian(
        shoebox.size(),
        1.0,
        [c - o for c, o in zip(centre[::-1], shoebox.offset())],
        [s / 8.0 for s in shoebox.size()],
    )
    if mask:
        shoebox.mask = create_mask(
            shoebox.size(),
            [c - o for c, o in zip(centre[::-1], shoebox.offset())],
            MaskCode.Valid | MaskCode.Foreground,
        )
    tot = 0
    mask_code = MaskCode.Valid | MaskCode.Foreground
    for i in range(len(shoebox.data)):
        if shoebox.mask[i] & mask_code == mask_code:
            tot += shoebox.data[i]
    if tot > 0:
        shoebox.data *= intensity / tot
    return shoebox


def create_mask(size, x0, value):
    from scitbx.array_family import flex

    mask = flex.int(flex.grid(size), 0)
    rad = min(s - c for s, c in zip(size, x0))
    for k in range(size[0]):
        for j in range(size[1]):
            for i in range(size[2]):
                d = math.sqrt((j - x0[1]) ** 2 + (i - x0[2]) ** 2)
                if d < rad:
                    mask[k, j, i] = value
    return mask


def evaluate_gaussian(x, a, x0, sx):
    assert len(x) == len(x0)
    assert len(x) == len(sx)

    g = 0.0
    for xi, x0i, sxi in zip(x, x0, sx):
        g += (xi - x0i) ** 2 / (2.0 * sxi**2)

    return a * math.exp(-g)


def gaussian(size, a, x0, sx):
    from dials.array_family import flex

    result = flex.real(flex.grid(size))

    index = [0] * len(size)
    while True:
        result[index] = evaluate_gaussian(index, a, x0, sx)
        for j in range(len(size)):
            index[j] += 1
            if index[j] < size[j]:
                break
            index[j] = 0
            if j == len(size) - 1:
                return result


def test_allocate():
    for i in range(10):
        x0 = random.randint(0, 1000)
        y0 = random.randint(0, 1000)
        z0 = random.randint(0, 1000)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox = Shoebox((x0, x1, y0, y1, z0, z1))
        shoebox.allocate()
        assert shoebox.data.all() == (z1 - z0, y1 - y0, x1 - x0)
        assert shoebox.mask.all() == (z1 - z0, y1 - y0, x1 - x0)
        shoebox.deallocate()
        assert shoebox.data.all() == (0, 0, 0)
        assert shoebox.mask.all() == (0, 0, 0)


def test_offset():
    for i in range(10):
        x0 = random.randint(0, 1000)
        y0 = random.randint(0, 1000)
        z0 = random.randint(0, 1000)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox = Shoebox((x0, x1, y0, y1, z0, z1))
        assert shoebox.xoffset() == x0
        assert shoebox.yoffset() == y0
        assert shoebox.zoffset() == z0
        assert shoebox.offset() == (z0, y0, x0)


def test_size():
    for i in range(10):
        x0 = random.randint(0, 1000)
        y0 = random.randint(0, 1000)
        z0 = random.randint(0, 1000)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox = Shoebox((x0, x1, y0, y1, z0, z1))
        assert shoebox.xsize() == x1 - x0
        assert shoebox.ysize() == y1 - y0
        assert shoebox.zsize() == z1 - z0
        assert shoebox.size() == (z1 - z0, y1 - y0, x1 - x0)


def test_consistent():
    from dials.array_family import flex

    for i in range(1000):
        x0 = random.randint(0, 1000)
        y0 = random.randint(0, 1000)
        z0 = random.randint(0, 1000)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0
        try:
            shoebox = Shoebox((x0, x1, y0, y1, z0, z1))
            assert not shoebox.is_consistent()
            shoebox.allocate()
            assert shoebox.is_consistent()
            shoebox.data = flex.real(flex.grid(20, 20, 20))
            assert not shoebox.is_consistent()
            shoebox.deallocate()
            assert not shoebox.is_consistent()
        except Exception:
            print(x0, y0, z0, x1, y1, z1)
            raise


def test_is_bbox_within_image_volume():
    isize = (1000, 1000)
    srange = (0, 100)

    shoebox = Shoebox((10, 20, 10, 20, 10, 20))
    assert shoebox.is_bbox_within_image_volume(isize, srange)
    shoebox = Shoebox((-10, 20, 10, 20, 10, 20))
    assert not shoebox.is_bbox_within_image_volume(isize, srange)
    shoebox = Shoebox((10, 20, -10, 20, 10, 20))
    assert not shoebox.is_bbox_within_image_volume(isize, srange)
    shoebox = Shoebox((10, 20, 10, 20, -10, 20))
    assert not shoebox.is_bbox_within_image_volume(isize, srange)
    shoebox = Shoebox((10, 1020, 10, 20, 10, 20))
    assert not shoebox.is_bbox_within_image_volume(isize, srange)
    shoebox = Shoebox((10, 20, 10, 1020, 10, 20))
    assert not shoebox.is_bbox_within_image_volume(isize, srange)
    shoebox = Shoebox((10, 20, 10, 20, 10, 1020))
    assert not shoebox.is_bbox_within_image_volume(isize, srange)


def test_does_bbox_contain_bad_pixels():
    from scitbx.array_family import flex

    mask = flex.bool(flex.grid(100, 100), True)
    for j in range(100):
        for i in range(40, 60):
            mask[j, i] = False
            mask[i, j] = False

    for i in range(1000):
        x0 = random.randint(0, 90)
        y0 = random.randint(0, 90)
        z0 = random.randint(0, 90)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox = Shoebox((x0, x1, y0, y1, z0, z1))

        res1 = shoebox.does_bbox_contain_bad_pixels(mask)
        res2 = False
        if x0 >= 40 and x0 < 60:
            res2 = True
        if x1 > 40 and x1 <= 60:
            res2 = True
        if y0 >= 40 and y0 < 60:
            res2 = True
        if y1 > 40 and y1 <= 60:
            res2 = True

        assert res1 == res2


def test_count_mask_values():
    for i in range(10):
        x0 = random.randint(0, 90)
        y0 = random.randint(0, 90)
        z0 = random.randint(0, 90)
        x1 = random.randint(1, 10) + x0
        y1 = random.randint(1, 10) + y0
        z1 = random.randint(1, 10) + z0

        shoebox = Shoebox((x0, x1, y0, y1, z0, z1))
        shoebox.allocate()
        maxnum = len(shoebox.mask)
        num = random.randint(1, maxnum)
        indices = random.sample(list(range(maxnum)), num)
        value = 1 << 2
        for i in indices:
            shoebox.mask[i] = value

        assert shoebox.count_mask_values(value) == num


def test_centroid_all():
    for shoebox, (XC, I) in random_shoeboxes(10):
        centroid = shoebox.centroid_all()
        assert shoebox.is_consistent()
        assert abs(matrix.col(centroid.px.position) - matrix.col(XC)) < 1.0


def test_centroid_masked():
    for shoebox, (XC, I) in random_shoeboxes(10):
        centroid = shoebox.centroid_masked(1 << 0)
        assert shoebox.is_consistent()
        assert abs(matrix.col(centroid.px.position) - matrix.col(XC)) < 1.0


def test_summed_intensity():
    for shoebox, (XC, I) in random_shoeboxes(10):
        intensity = shoebox.summed_intensity()
        assert shoebox.is_consistent()
        assert abs(intensity.observed.value - I) < 1e-1


def test_flatten():
    from dials.algorithms.shoebox import MaskCode
    from dials.array_family import flex

    for shoebox, (XC, I) in random_shoeboxes(10, mask=True):
        assert not shoebox.flat
        zs = shoebox.zsize()
        ys = shoebox.ysize()
        xs = shoebox.xsize()
        expected_data = flex.real(flex.grid(1, ys, xs), 0)
        expected_mask = flex.int(flex.grid(1, ys, xs), 0)
        for k in range(zs):
            for j in range(ys):
                for i in range(xs):
                    expected_data[0, j, i] += shoebox.data[k, j, i]
                    expected_mask[0, j, i] |= shoebox.mask[k, j, i]
                    if not (expected_mask[0, j, i] & MaskCode.Valid) or not (
                        shoebox.mask[k, j, i] & MaskCode.Valid
                    ):
                        expected_mask[0, j, i] &= ~MaskCode.Valid
        shoebox.flatten()
        diff = expected_data.as_double() - shoebox.data.as_double()
        max_diff = flex.max(flex.abs(diff))
        assert max_diff < 1e-7
        assert expected_mask.all_eq(shoebox.mask)
        assert shoebox.flat
        assert shoebox.is_consistent()


def test_all_foreground_valid():
    from .all_foreground_valid_data import data

    shoeboxes = pickle.loads(bytes(data, encoding="latin-1"), encoding="bytes")
    for i, shoebox in enumerate(shoeboxes):
        if i < 4:
            assert not shoebox.all_foreground_valid()
        else:
            assert shoebox.all_foreground_valid()



from __future__ import annotations

import logging

from libtbx.phil import parse

from dials.array_family import flex
from dials.model.experiment.profile import ProfileModelExt

logger = logging.getLogger(__name__)

phil_scope = parse(
    """

  gaussian_rs
  {
    scan_varying = False
        .type = bool
        .help = "Calculate a scan varying model"

    min_spots
      .help = "if (total_reflections > overall or reflections_per_degree >"
              "per_degree) then do the profile modelling."
    {
      overall = 50
        .type = int(value_min=0)
        .help = "The minimum number of spots needed to do the profile modelling"

      per_degree = 20
        .type = int(value_min=0)
        .help = "The minimum number of spots needed to do the profile modelling"
    }

    sigma_m_algorithm = basic *extended
      .type = choice
      .help = "The algorithm to compute mosaicity"

    centroid_definition = com *s1
      .type = choice
      .help = "The centroid to use as beam divergence (centre of mass or s1)"

    parameters {

      n_sigma = 3.0
        .type = float(value_min=0)
        .help = "Sigma multiplier for shoebox"

      sigma_b = None
        .type = float(value_min=0)
        .help = "Override the sigma_b value (degrees)"

      sigma_m = None
        .type = float(value_min=0)
        .help = "Override the sigma_m value (degrees)"
    }

    filter
    {
      min_zeta = 0.05
        .type = float
        .help = "Filter reflections by min zeta"
    }

    fitting {

      scan_step = 5
        .type = float
        .help = "Space between profiles in degrees"

      grid_size = 5
        .type = int
        .help = "The size of the profile grid."

      threshold = 0.02
        .type = float
        .help = "The threshold to use in reference profile"

      grid_method = single *regular_grid circular_grid spherical_grid
        .type = choice
        .help = "Select the profile grid method"

      fit_method = *reciprocal_space detector_space
        .type = choice
        .help = "The fitting method"

      detector_space {

        deconvolution = False
          .type = bool
          .help = "Do deconvolution in detector space"
      }

    }
  }
"""
)


class Model(ProfileModelExt):
    def __init__(self, params, n_sigma, sigma_b, sigma_m, deg=False):
        """Initialise with the parameters."""
        from math import pi

        self.params = params
        self._n_sigma = n_sigma
        if deg:
            self._sigma_b = sigma_b * pi / 180.0
            self._sigma_m = sigma_m * pi / 180.0
        else:
            self._sigma_b = sigma_b
            self._sigma_m = sigma_m
        assert self._n_sigma > 0
        if isinstance(self._sigma_b, flex.double):
            self._scan_varying = True
        else:
            self._scan_varying = False
        if self._scan_varying:
            assert self._sigma_b.all_gt(0)
            assert self._sigma_m.all_gt(0)
            assert len(self._sigma_b) == len(self._sigma_m)
        else:
            assert self._sigma_b > 0
            assert self._sigma_m >= 0
            assert self._n_sigma > 0

    @classmethod
    def from_dict(cls, obj):
        """Convert the profile model from a dictionary."""
        if obj["__id__"] != "gaussian_rs":
            raise RuntimeError(f"expected __id__ gaussian_rs, got {obj['__id__']}")
        n_sigma = obj["n_sigma"]
        sigma_b = obj["sigma_b"]
        sigma_m = obj["sigma_m"]
        if isinstance(sigma_b, list):
            assert len(sigma_b) == len(sigma_m)
            sigma_b = flex.double(sigma_b)
            sigma_m = flex.double(sigma_m)
        return cls(None, n_sigma, sigma_b, sigma_m, deg=True)

    def to_dict(self):
        """Convert the model to a dictionary."""
        n_sigma = self.n_sigma()
        if self._scan_varying:
            sigma_b = list(self.sigma_b(deg=True))
            sigma_m = list(self.sigma_m(deg=True))
        else:
            sigma_b = self.sigma_b(deg=True)
            sigma_m = self.sigma_m(deg=True)
        return {
            "__id__": "gaussian_rs",
            "n_sigma": n_sigma,
            "sigma_b": sigma_b,
            "sigma_m": sigma_m,
        }

    def sigma_b(self, index=None, deg=True):
        """Return sigma_b."""
        from math import pi

        if index is None:
            sigma_b = self._sigma_b
        else:
            sigma_b = self._sigma_b[index]
        if deg:
            return sigma_b * 180.0 / pi
        return sigma_b

    def sigma_m(self, index=None, deg=True):
        """Return sigma_m."""
        from math import pi

        if index is None:
            sigma_m = self._sigma_m
        else:
            sigma_m = self._sigma_m[index]
        if deg:
            return sigma_m * 180.0 / pi
        return sigma_m

    def n_sigma(self):
        """The number of sigmas."""
        return self._n_sigma

    def delta_b(self, index=None, deg=True):
        """Return delta_b."""
        return self.sigma_b(index, deg) * self.n_sigma()

    def delta_m(self, index=None, deg=True):
        """Return delta_m."""
        return self.sigma_m(index, deg) * self.n_sigma()

    def is_scan_varying(self):
        """Return whether scan varying."""
        return self._scan_varying

    def num_scan_points(self):
        """Return number of scan points."""
        if self._scan_varying:
            assert len(self._sigma_m) == len(self._sigma_b)
            return len(self._sigma_m)
        return 1

    @classmethod
    def create(
        cls,
        params,
        reflections,
        crystal,
        beam,
        detector,
        goniometer=None,
        scan=None,
        profile=None,
    ):
        """
        Create the profile model from data.

        :param params: The phil parameters
        :param reflections: The reflections
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        :return: An instance of the profile model
        """
        if reflections is not None:
            model = cls.create_from_reflections(
                params, reflections, crystal, beam, detector, goniometer, scan, profile
            )
        else:
            model = cls.create_from_parameters(
                params, reflections, crystal, beam, detector, goniometer, scan, profile
            )
        return model

    @classmethod
    def create_from_parameters(
        cls,
        params,
        reflections,
        crystal,
        beam,
        detector,
        goniometer=None,
        scan=None,
        profile=None,
    ):
        """
        Create the profile model from parameters.

        :param params: The phil parameters
        :param reflections: The reflections
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        :return: An instance of the profile model
        """
        if profile is not None:
            sigma_b = profile.sigma_b()
            sigma_m = profile.sigma_m()
        else:
            sigma_b = None
            sigma_m = None
        if params.gaussian_rs.parameters.sigma_b is not None:
            sigma_b = params.gaussian_rs.parameters.sigma_b
        if params.gaussian_rs.parameters.sigma_m is not None:
            sigma_m = params.gaussian_rs.parameters.sigma_m
        n_sigma = params.gaussian_rs.parameters.n_sigma
        if sigma_m is None or sigma_b is None:
            raise RuntimeError(
                "Not enough information to set profile parameters. "
                "These can be calculated from a reflection file (if provided) "
                "or set using the parameters sigma_b and sigma_m."
            )
        logger.info("Creating profile model with parameters:")
        logger.info("  sigma_b: %f", sigma_b)
        logger.info("  sigma_m: %f", sigma_m)
        return cls(
            params=params, n_sigma=n_sigma, sigma_b=sigma_b, sigma_m=sigma_m, deg=True
        )

    @classmethod
    def create_from_reflections(
        cls,
        params,
        reflections,
        crystal,
        beam,
        detector,
        goniometer=None,
        scan=None,
        profile=None,
    ):
        """
        Create the profile model from data.

        :param params: The phil parameters
        :param reflections: The reflections
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        :return: An instance of the profile model
        """
        from dials.algorithms.profile_model.gaussian_rs.calculator import (
            ProfileModelCalculator,
            ScanVaryingProfileModelCalculator,
        )

        # Check the number of spots
        if not len(reflections) >= params.gaussian_rs.min_spots.overall:
            if scan is not None:
                num_degrees = scan.get_num_images() * scan.get_oscillation()[1]
                spots_per_degree = len(reflections) / num_degrees
                if not spots_per_degree >= params.gaussian_rs.min_spots.per_degree:
                    raise RuntimeError(
                        """
            Too few reflections for profile modelling:
              need %d per degree or  %d in total
              got  %d per degree and %d in total

            The default values may be conservative. Using fewer spots may still work.
            A solution may be to try changing parameters:
              profile.gaussian_rs.min_spots.per_degree=%d
              profile.gaussian_rs.min_spots.overall=%d
            """
                        % (
                            params.gaussian_rs.min_spots.per_degree,
                            params.gaussian_rs.min_spots.overall,
                            spots_per_degree,
                            len(reflections),
                            params.gaussian_rs.min_spots.per_degree,
                            params.gaussian_rs.min_spots.overall,
                        )
                    )
            else:
                raise RuntimeError(
                    """
          Too few reflections for profile modelling:
            expected >= %d, got %d

          The default values may be conservative. Using fewer spots may still work.
          A solution may be to try changing parameters:
            profile.gaussian_rs.min_spots.per_degree=%d
            profile.gaussian_rs.min_spots.overall=%d
          """
                    % (
                        params.gaussian_rs.min_spots.overall,
                        len(reflections),
                        params.gaussian_rs.min_spots.per_degree,
                        params.gaussian_rs.min_spots.overall,
                    )
                )
        if not len(reflections) > 0:
            raise RuntimeError(
                """
        No reflections remaining for profile modelling.
      """
            )

        # Check the override parameters
        if [
            params.gaussian_rs.parameters.sigma_b,
            params.gaussian_rs.parameters.sigma_m,
        ].count(None) == 1:
            raise RuntimeError("sigma_b and sigma_m parameters must both be set")
        if (
            params.gaussian_rs.parameters.sigma_b is not None
            and params.gaussian_rs.parameters.sigma_m is not None
        ):
            return cls(
                params=params,
                n_sigma=params.gaussian_rs.parameters.n_sigma,
                sigma_b=params.gaussian_rs.parameters.sigma_b,
                sigma_m=params.gaussian_rs.parameters.sigma_m,
                deg=True,
            )

        if not params.gaussian_rs.scan_varying:
            Calculator = ProfileModelCalculator
        else:
            Calculator = ScanVaryingProfileModelCalculator
        calculator = Calculator(
            reflections,
            crystal,
            beam,
            detector,
            goniometer,
            scan,
            params.gaussian_rs.filter.min_zeta,
            algorithm=params.gaussian_rs.sigma_m_algorithm,
            centroid_definition=params.gaussian_rs.centroid_definition,
        )
        return cls(
            params=params,
            n_sigma=params.gaussian_rs.parameters.n_sigma,
            sigma_b=calculator.sigma_b(),
            sigma_m=calculator.sigma_m(),
        )

    def predict_reflections(
        self,
        imageset,
        crystal,
        beam,
        detector,
        goniometer=None,
        scan=None,
        dmin=None,
        dmax=None,
        margin=1,
        force_static=False,
        padding=0,
        **kwargs,
    ):
        """
        Given an experiment, predict the reflections.

        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        from dxtbx.model.experiment_list import Experiment

        from dials.algorithms.spot_prediction.reflection_predictor import (
            ReflectionPredictor,
        )

        predict = ReflectionPredictor(
            Experiment(
                imageset=imageset,
                crystal=crystal,
                beam=beam,
                detector=detector,
                goniometer=goniometer,
                scan=scan,
            ),
            dmin=dmin,
            dmax=dmax,
            margin=margin,
            force_static=force_static,
            padding=padding,
        )
        return predict()

    def compute_partiality(
        self, reflections, crystal, beam, detector, goniometer=None, scan=None, **kwargs
    ):
        """
        Given an experiment and list of reflections, compute the partiality of the
        reflections

        :param reflections: The reflection table
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        from dials.algorithms.profile_model.gaussian_rs import PartialityCalculator

        # Create the partiality calculator
        calculate = PartialityCalculator(
            crystal, beam, detector, goniometer, scan, self._sigma_m
        )

        # Compute the partiality
        partiality = calculate(
            reflections["s1"], reflections["xyzcal.px"].parts()[2], reflections["bbox"]
        )

        # Return the partiality
        return partiality

    def compute_bbox(
        self,
        reflections,
        crystal,
        beam,
        detector,
        goniometer=None,
        scan=None,
        sigma_b_multiplier=2.0,
        **kwargs,
    ):
        """Given an experiment and list of reflections, compute the
        bounding box of the reflections on the detector (and image frames).

        :param reflections: The reflection table
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        from dials.algorithms.profile_model.gaussian_rs import BBoxCalculator

        # Check the input
        assert sigma_b_multiplier >= 1.0

        # Compute the size in reciprocal space. Add a sigma_b multiplier to enlarge
        # the region of background in the shoebox
        delta_b = self._n_sigma * self._sigma_b * sigma_b_multiplier
        delta_m = self._n_sigma * self._sigma_m

        # Create the bbox calculator
        calculate = BBoxCalculator(
            crystal, beam, detector, goniometer, scan, delta_b, delta_m
        )

        # Calculate the bounding boxes of all the reflections
        bbox = calculate(
            reflections["s1"], reflections["xyzcal.px"].parts()[2], reflections["panel"]
        )

        # Return the bounding boxes
        return bbox

    def compute_mask(
        self,
        reflections,
        crystal,
        beam,
        detector,
        goniometer=None,
        scan=None,
        image_volume=None,
        **kwargs,
    ):
        """
        Given an experiment and list of reflections, compute the
        foreground/background mask of the reflections.

        :param reflections: The reflection table
        :param crystal: The crystal model
        :param beam: The beam model
        :param detector: The detector model
        :param goniometer: The goniometer model
        :param scan: The scan model
        """
        from dials.algorithms.profile_model.gaussian_rs import MaskCalculator

        # Compute the size in reciprocal space. Add a sigma_b multiplier to enlarge
        # the region of background in the shoebox
        delta_b = self._n_sigma * self._sigma_b
        delta_m = self._n_sigma * self._sigma_m

        # Create the mask calculator
        mask_foreground = MaskCalculator(
            crystal, beam, detector, goniometer, scan, delta_b, delta_m
        )

        # Mask the foreground region
        if image_volume is None:
            return mask_foreground(
                reflections["shoebox"],
                reflections["s1"],
                reflections["xyzcal.px"].parts()[2],
                reflections["panel"],
            )
        else:
            return mask_foreground(
                image_volume,
                reflections["bbox"],
                reflections["s1"],
                reflections["xyzcal.px"].parts()[2],
                reflections["panel"],
            )

    def fitting_class(self):
        """
        Get the profile fitting algorithm associated with this profile model

        :return: The profile fitting class
        """

        # Check input
        if self._scan_varying:
            if not self.sigma_m().all_gt(0):
                return None
        else:
            if self.sigma_m() <= 0:
                return None

        # Define a function to create the fitting class
        def wrapper(experiment):
            from math import ceil

            from dials.algorithms.profile_model.gaussian_rs import (
                GaussianRSProfileModeller,
            )

            # Return if no scan or gonio
            if (
                experiment.scan is None
                or experiment.goniometer is None
                or experiment.scan.is_still()
            ):
                return None

            # Compute the scan step
            phi0, phi1 = experiment.scan.get_oscillation_range(deg=True)
            assert phi1 > phi0
            phi_range = phi1 - phi0
            num_scan_points = int(
                ceil(phi_range / self.params.gaussian_rs.fitting.scan_step)
            )
            assert num_scan_points > 0

            # Create the grid method
            GridMethod = GaussianRSProfileModeller.GridMethod
            FitMethod = GaussianRSProfileModeller.FitMethod
            grid_method = int(
                GridMethod.names[self.params.gaussian_rs.fitting.grid_method].real
            )
            fit_method = int(
                FitMethod.names[self.params.gaussian_rs.fitting.fit_method].real
            )

            if self._scan_varying:
                sigma_b = flex.mean(self.sigma_b(deg=False))
                sigma_m = flex.mean(self.sigma_m(deg=False))
            else:
                sigma_b = self.sigma_b(deg=False)
                sigma_m = self.sigma_m(deg=False)

            # Create the modeller
            return GaussianRSProfileModeller(
                experiment.beam,
                experiment.detector,
                experiment.goniometer,
                experiment.scan,
                sigma_b,
                sigma_m,
                self.n_sigma() * 1.5,
                self.params.gaussian_rs.fitting.grid_size,
                num_scan_points,
                self.params.gaussian_rs.fitting.threshold,
                grid_method,
                fit_method,
            )

        # Return the wrapper function
        return wrapper

    def __str__(self):
        return "\n".join(
            [
                "Profile model:",
                "    type: gaussian_rs",
                f"    delta_b (sigma_b): {self.delta_b():f} ({self.sigma_b():f})",
                f"    delta_m (sigma_m): {self.delta_m():f} ({self.sigma_m():f})",
            ]
        )



from __future__ import annotations


def test_after_import_extensions():
    import dials.extensions
    from dials.extensions.dispersion_spotfinder_threshold_ext import (
        DispersionSpotFinderThresholdExt,
    )

    # Check the expected extensions are present for each interface
    from dials.extensions.gaussian_rs_profile_model_ext import GaussianRSProfileModelExt
    from dials.extensions.null_background_ext import NullBackgroundExt
    from dials.extensions.simple_background_ext import SimpleBackgroundExt
    from dials.extensions.simple_centroid_ext import SimpleCentroidExt

    extensions = dials.extensions.ProfileModel.extensions()
    assert GaussianRSProfileModelExt in extensions
    extensions = dials.extensions.SpotFinderThreshold.extensions()
    assert DispersionSpotFinderThresholdExt in extensions
    extensions = dials.extensions.Centroid.extensions()
    assert SimpleCentroidExt in extensions
    extensions = dials.extensions.Background.extensions()
    assert NullBackgroundExt in extensions
    assert SimpleBackgroundExt in extensions

    # Check phil scope
    phil_scope = dials.extensions.ProfileModel.phil_scope()
    assert phil_scope
    phil_scope = dials.extensions.SpotFinderThreshold.phil_scope()
    assert phil_scope
    phil_scope = dials.extensions.Centroid.phil_scope()
    assert phil_scope
    phil_scope = dials.extensions.Background.phil_scope()
    assert phil_scope



#!/usr/bin/env python
#
# dials.ssx_integrate.py
#
#  Copyright (C) 2022 Diamond Light Source
#
#  Author: James Beilsten-Edmands
#
#  This code is distributed under the BSD license, a copy of which is
#  included in the root directory of this package.
"""
This program rums profile modelling and integration on indexed results from a
still sequence i.e. SSX data. This script uses parts of the regular DIALS
integration code, using either the ellipsoid or stills integrator algorithms.

The ellipsoid algorithm refines the unit cell, orientation and a 3D ellipsoidal
mosaicity parameterisation for each crystal, by assessing the pixel-intensity
distribution of the strong spots. The integrated data are saved in batches to
hep with memory management. A html output report is generated, showing integration
and clutering statistics.

Further program documentation can be found at dials.github.io/ssx_processing_guide.html

Usage:
    dials.ssx_integrate indexed.expt indexed.refl
    dials.ssx_integrate refined.expt refined.refl
    dials.ssx_integrate indexed.expt indexed.refl algorithm=stills
"""

from __future__ import annotations

import copy
import json
import logging
import pathlib
from dataclasses import dataclass
from multiprocessing import Pool
from typing import Any

import iotbx.phil
from cctbx import crystal
from dxtbx.model import Experiment, ExperimentList
from libtbx import Auto
from libtbx.introspection import number_of_processors
from libtbx.utils import Sorry

from dials.algorithms.indexing.ssx.analysis import report_on_crystal_clusters
from dials.algorithms.indexing.ssx.processing import manage_loggers
from dials.algorithms.integration.ssx.ellipsoid_integrate import (
    EllipsoidIntegrator,
    EllipsoidOutputAggregator,
)
from dials.algorithms.integration.ssx.ssx_integrate import (
    OutputAggregator,
    SimpleIntegrator,
    generate_html_report,
)
from dials.algorithms.integration.ssx.stills_integrate import StillsIntegrator
from dials.array_family import flex
from dials.command_line.combine_experiments import CombineWithReference
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, flatten_experiments, flatten_reflections
from dials.util.version import dials_version

try:
    from typing import List, Type
except ImportError:
    pass

logger = logging.getLogger("dials.ssx_integrate")

# Create the phil scope
phil_scope = iotbx.phil.parse(
    """
  algorithm = *ellipsoid stills
    .type = choice
  nproc=Auto
    .type = int
  image_range = None
    .type = str
    .help = "Only process this image range. Input in the format start:end."
            "This range is an inclusive range."
  output {
    batch_size = 50
      .type = int
      .help = "Number of images to save in each output file"
    log = "dials.ssx_integrate.log"
      .type = str
    html = "dials.ssx_integrate.html"
      .type = str
    json = None
      .type = str
    nuggets = None
      .type = path
      .help = "Specify a directory to which a per-image summary json will be saved"
              "during processing, as each image is integrated, to enable live monitoring."
  }

  ellipsoid {
    include scope dials.algorithms.profile_model.ellipsoid.algorithm.ellipsoid_algorithm_phil_scope
  }

  include scope dials.algorithms.integration.integrator.phil_scope
  include scope dials.algorithms.profile_model.factory.phil_scope
  include scope dials.algorithms.spot_prediction.reflection_predictor.phil_scope
  include scope dials.algorithms.integration.stills_significance_filter.phil_scope
  include scope dials.algorithms.integration.kapton_correction.absorption_phil_scope


  debug {
    output {
      shoeboxes = False
        .type = bool
    }
  }

""",
    process_includes=True,
)

phil_overrides = phil_scope.fetch(
    source=iotbx.phil.parse(
        """\
profile {
    gaussian_rs {
        min_spots {
            overall=0
        }
    }
    fitting = False
    ellipsoid {
        refinement {
            n_cycles = 1
        }
    }
}
integration {
    background {
        simple {
            outlier {
                algorithm = Null
            }
        }
    }
}
"""
    )
)
working_phil = phil_scope.fetch(sources=[phil_overrides])

working_phil.adopt_scope(
    iotbx.phil.parse(
        """
    individual_log_verbosity = 1
    .type =int
"""
    )
)

loggers_to_disable = ["dials", "dials.array_family.flex_ext"]

loggers_to_disable_for_stills = loggers_to_disable + [
    "dials.algorithms.integration.integrator",
    "dials.algorithms.profile_model.gaussian_rs.calculator",
    "dials.command_line.integrate",
    "dials.algorithms.spot_prediction.reflection_predictor",
]


def process_one_image(experiment, table, params, integrator_class):
    collect_data = params.output.html or params.output.json
    integrator = integrator_class(params, collect_data)
    try:
        experiment, table, collector = integrator.run(experiment, table)
    except RuntimeError as e:
        logger.info(f"Processing failed due to error: {e}")
        return (None, None, None)
    else:
        return experiment, table, collector


def setup(reflections, params):
    # calculate the batches for processing
    if params.image_range:
        if not len(params.image_range.split(":")) == 2:
            raise ValueError("Image range must be given in the form first:last")
        first, last = params.image_range.split(":")
        try:
            first = int(first)
        except ValueError:
            raise ValueError(f"Issue interpreting {first} as an integer")
        try:
            last = int(last)
        except ValueError:
            raise ValueError(f"Issue interpreting {last} as an integer")
        lowest_index = max(0, first - 1)
        highest_image = min(last, len(reflections))
        batches = list(range(lowest_index, highest_image, params.output.batch_size))
        batches.append(highest_image)
    else:
        batches = list(range(0, len(reflections), params.output.batch_size))
        batches.append(len(reflections))

    # Note, memory processing logic can go here
    if params.nproc is Auto:
        params.nproc = number_of_processors(return_value_if_unknown=1)
    logger.info(f"Using {params.nproc} processes for integration")

    # aggregate some output for json, html etc
    if params.algorithm == "ellipsoid":
        process = EllipsoidIntegrator
        aggregator = EllipsoidOutputAggregator()
    elif params.algorithm == "stills":
        process = StillsIntegrator
        aggregator = OutputAggregator()
    else:
        raise ValueError("Invalid algorithm choice")

    configuration = {
        "process": process,
        "aggregator": aggregator,
        "params": params,
        "loggers_to_disable": (
            loggers_to_disable
            if params.algorithm == "ellipsoid"
            else loggers_to_disable_for_stills
        ),
    }

    return batches, configuration


@dataclass
class InputToIntegrate:
    integrator_class: Type[SimpleIntegrator]
    experiment: Experiment
    table: flex.reflection_table
    params: Any
    crystalno: int


@dataclass
class IntegrationResult:
    experiment: Experiment
    table: flex.reflection_table
    collector: Any
    crystalno: int


def wrap_integrate_one(input_to_integrate: InputToIntegrate):
    expt, refls, collector = process_one_image(
        input_to_integrate.experiment,
        input_to_integrate.table,
        input_to_integrate.params,
        input_to_integrate.integrator_class,
    )

    result = IntegrationResult(expt, refls, collector, input_to_integrate.crystalno)
    if expt and refls:
        if not input_to_integrate.params.debug.output.shoeboxes:
            del result.table["shoebox"]
        logger.info(f"Processed crystal {input_to_integrate.crystalno}")
    if input_to_integrate.params.output.nuggets:
        img = input_to_integrate.experiment.imageset.get_image_identifier(0).split("/")[
            -1
        ]
        msg = {
            "crystal_no": input_to_integrate.crystalno,
            "n_integrated": 0,
            "i_over_sigma_overall": 0,
            "image": img,
        }
        if expt and refls:
            msg["n_integrated"] = collector.data["n_integrated"]
            msg["i_over_sigma_overall"] = round(
                collector.data["i_over_sigma_overall"], 2
            )

        with open(
            input_to_integrate.params.output.nuggets
            / f"nugget_integrated_{input_to_integrate.crystalno}.json",
            "w",
        ) as f:
            f.write(json.dumps(msg))
    return result


def process_batch(sub_tables, sub_expts, configuration, batch_offset=0):

    # create iterable
    input_iterable: List[InputToIntegrate] = []
    for i, (table, expt) in enumerate(zip(sub_tables, sub_expts)):
        input_iterable.append(
            InputToIntegrate(
                configuration["process"],
                expt,
                table,
                configuration["params"],
                i + 1 + batch_offset,
            )
        )

    with manage_loggers(
        configuration["params"].individual_log_verbosity,
        configuration["loggers_to_disable"],
    ):
        if configuration["params"].nproc > 1:
            with Pool(configuration["params"].nproc) as pool:
                results: List[IntegrationResult] = pool.map(
                    wrap_integrate_one, input_iterable
                )
        else:
            results: List[IntegrationResult] = [
                wrap_integrate_one(i) for i in input_iterable
            ]

    # then join
    integrated_reflections = flex.reflection_table()
    integrated_experiments = []

    n_integrated = 0
    for result in results:
        if result.table:
            ids_map = dict(result.table.experiment_identifiers())
            del result.table.experiment_identifiers()[list(ids_map.keys())[0]]
            result.table["id"] = flex.int(result.table.size(), n_integrated)
            result.table.experiment_identifiers()[n_integrated] = list(
                ids_map.values()
            )[0]
            n_integrated += 1
            integrated_reflections.extend(result.table)
            integrated_experiments.append(result.experiment)
            configuration["aggregator"].add_dataset(result.collector, result.crystalno)

    integrated_experiments = ExperimentList(integrated_experiments)
    integrated_reflections.assert_experiment_identifiers_are_consistent(
        integrated_experiments
    )
    return integrated_experiments, integrated_reflections


def run_integration(reflections, experiments, params):
    assert len(reflections) == len(experiments)
    if params.output.nuggets:
        params.output.nuggets = pathlib.Path(params.output.nuggets)
        if not params.output.nuggets.is_dir():
            logger.warning(
                "output.nuggets not recognised as a valid directory path, no nuggets will be output"
            )
            params.output.nuggets = None
    batches, configuration = setup(reflections, params)

    # now process each batch, and do parallel processing within a batch
    for i, b in enumerate(batches[:-1]):
        end_ = batches[i + 1]
        logger.info(f"Processing images {b+1} to {end_}")
        sub_tables = reflections[b:end_]
        sub_expts = experiments[b:end_]

        integrated_experiments, integrated_reflections = process_batch(
            sub_tables, sub_expts, configuration, batch_offset=b
        )
        yield integrated_experiments, integrated_reflections, configuration[
            "aggregator"
        ]


@show_mail_handle_errors()
def run(args: List[str] = None, phil=working_phil) -> None:
    """
    Run dials.ssx_integrate from the command-line.

    This program takes an indexed experiment list and reflection table and
    performs parallelised integration for synchrotron serial crystallography
    experiments. The programs acts as a wrapper to run one of two algorithms,
    the stills integrator or the 'ellipsoid' integrator (which uses a generalised
    ellipsoidal profile model). Analysis statistics are captured and output as
    a html report, while the output data are saved in batches for memory
    management.
    """

    parser = ArgumentParser(
        usage="dials.ssx_integrate indexed.expt indexed.refl [options]",
        phil=phil,
        epilog=__doc__,
        read_experiments=True,
        read_reflections=True,
    )
    # Check the number of arguments is correct

    # Parse the command line
    params, options = parser.parse_args(args=args, show_diff_phil=False)
    reflections = flatten_reflections(params.input.reflections)
    experiments = flatten_experiments(params.input.experiments)

    if len(reflections) == 0 or len(experiments) == 0:
        parser.print_help()
        return

    if len(reflections) != 1:
        raise Sorry(
            "Only a single reflection table file can be input (this can be a multi-still table)"
        )

    # Configure logging
    log.config(verbosity=options.verbose, logfile=params.output.log)
    params.individual_log_verbosity = options.verbose
    logger.info(dials_version())

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    ## FIXME - experiment identifiers approach wont work if input strong.refl and refined.expt
    # for now - check image path and update identifiers to that of refined.expt?
    if len(set(reflections[0]["id"]).difference({-1})) > 1:
        logger.info("Attempting to split multi-still reflection table")
        reflections = reflections[0].split_by_experiment_id()
        if not (len(reflections) == len(experiments)):
            raise Sorry(
                "Unequal number of reflection tables and experiments after splitting"
            )

    integrated_crystal_symmetries = []

    for i, (int_expt, int_refl, aggregator) in enumerate(
        run_integration(reflections, experiments, params)
    ):
        # combine beam and detector models if not already
        if len(int_expt.detectors()) > 1 or len(int_expt.beams()) > 1:
            combine = CombineWithReference(
                detector=int_expt[0].detector, beam=int_expt[0].beam
            )
            elist = ExperimentList()
            for expt in int_expt:
                elist.append(combine(expt))
            int_expt = elist
        reflections_filename = f"integrated_{i+1}.refl"
        experiments_filename = f"integrated_{i+1}.expt"
        logger.info(f"Saving {int_refl.size()} reflections to {reflections_filename}")
        int_refl.as_file(reflections_filename)
        logger.info(f"Saving the experiments to {experiments_filename}")
        int_expt.as_file(experiments_filename)

        integrated_crystal_symmetries.extend(
            [
                crystal.symmetry(
                    unit_cell=copy.deepcopy(cryst.get_unit_cell()),
                    space_group=copy.deepcopy(cryst.get_space_group()),
                )
                for cryst in int_expt.crystals()
            ]
        )

    plots = {}
    cluster_plots, _ = report_on_crystal_clusters(
        integrated_crystal_symmetries,
        make_plots=(params.output.html or params.output.json),
    )

    if params.output.html or params.output.json:
        # now generate plots using the aggregated data.
        plots = aggregator.make_plots()
        plots.update(cluster_plots)

    if params.output.html and plots:
        logger.info(f"Writing html report to {params.output.html}")
        generate_html_report(plots, params.output.html)
    if params.output.json and plots:
        logger.info(f"Saving plot data in json format to {params.output.json}")
        with open(params.output.json, "w") as outfile:
            json.dump(plots, outfile, indent=2)

    logger.info(
        "Further program documentation can be found at dials.github.io/ssx_processing_guide.html"
    )


if __name__ == "__main__":
    run()



"""
Helper functions/classes for making HTML report and scaling summary output.
"""

from __future__ import annotations

import json
import logging
import math

from jinja2 import ChoiceLoader, Environment, PackageLoader
from orderedset import OrderedSet

from cctbx import uctbx
from dxtbx import flumpy
from scitbx.array_family import flex

from dials.algorithms.scaling.error_model.error_model import (
    calc_deltahl,
    calc_sigmaprime,
)
from dials.algorithms.scaling.error_model.error_model_target import (
    calculate_regression_x_y,
)
from dials.algorithms.scaling.model.model import (
    make_combined_plots,
    plot_scaling_models,
)
from dials.algorithms.scaling.plots import (
    error_model_variance_plot,
    error_regression_plot,
    normal_probability_plot,
    plot_outliers,
)
from dials.algorithms.scaling.scale_and_filter import make_scaling_filtering_plots
from dials.algorithms.scaling.scaling_library import (
    DialsMergingStatisticsError,
    merging_stats_from_scaled_array,
)
from dials.report.analysis import (
    make_merging_statistics_summary,
    reflection_tables_to_batch_dependent_properties,
    table_1_summary,
)
from dials.report.plots import (
    AnomalousPlotter,
    IntensityStatisticsPlots,
    ResolutionPlotsAndStats,
    i_over_sig_i_vs_batch_plot,
    i_over_sig_i_vs_i_plot,
    make_image_range_table,
    scale_rmerge_vs_batch_plot,
)
from dials.util import tabulate
from dials.util.batch_handling import batch_manager, get_image_ranges
from dials.util.exclude_images import get_valid_image_ranges
from dials.util.resolution_analysis import resolution_cc_half

logger = logging.getLogger("dials")


def assert_is_json_serialisable(thing, name, path=None):
    path = path or []
    if isinstance(thing, list):
        for n, element in enumerate(thing):
            assert_is_json_serialisable(element, name, path + [n])
    elif isinstance(thing, dict):
        for key, value in thing.items():
            assert_is_json_serialisable(value, name, path + [repr(key)])
    else:
        try:
            json.dumps(thing)
        except TypeError as e:
            raise TypeError(
                "JSON serialisation error '%s' for value '%s' type %s in %s%s"
                % (
                    e,
                    str(thing),
                    type(thing),
                    name,
                    "".join("[%s]" % step for step in path),
                )
            )


class ScalingSummaryContextManager:
    def __init__(self, script):
        self.script = script

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        print_scaling_summary(self.script)


def print_scaling_summary(script):
    """Log summary information after scaling."""
    logger.info(print_scaling_model_error_summary(script.experiments))
    valid_ranges = get_valid_image_ranges(script.experiments)
    image_ranges = get_image_ranges(script.experiments)
    msg = []
    for (img, valid, refl) in zip(image_ranges, valid_ranges, script.reflections):
        if valid:
            if len(valid) > 1 or valid[0][0] != img[0] or valid[-1][1] != img[1]:
                msg.append(
                    "Excluded images for experiment id: %s, image range: %s, limited range: %s"
                    % (
                        refl.experiment_identifiers().keys()[0],
                        list(img),
                        list(valid),
                    )
                )
    if msg:
        msg = ["Summary of image ranges removed:"] + msg
        logger.info("\n".join(msg))

    # report on partiality of dataset
    partials = flex.double()
    for r in script.reflections:
        if "partiality" in r:
            partials.extend(r["partiality"])
    not_full_sel = partials < 0.99
    not_zero_sel = partials > 0.01
    gt_half = partials > 0.5
    lt_half = partials < 0.5
    partial_gt_half_sel = not_full_sel & gt_half
    partial_lt_half_sel = not_zero_sel & lt_half
    logger.info("Summary of dataset partialities")
    header = ["Partiality (p)", "n_refl"]
    rows = [
        ["all reflections", str(partials.size())],
        ["p > 0.99", str(not_full_sel.count(False))],
        ["0.5 < p < 0.99", str(partial_gt_half_sel.count(True))],
        ["0.01 < p < 0.5", str(partial_lt_half_sel.count(True))],
        ["p < 0.01", str(not_zero_sel.count(False))],
    ]
    logger.info(tabulate(rows, header))
    logger.info(
        """
Reflections below a partiality_cutoff of %s are not considered for any
part of the scaling analysis or for the reporting of merging statistics.
Additionally, if applicable, only reflections with a min_partiality > %s
were considered for use when refining the scaling model.
""",
        script.params.cut_data.partiality_cutoff,
        script.params.reflection_selection.min_partiality,
    )
    stats = script.merging_statistics_result
    if stats:
        anom_stats, cut_stats, cut_anom_stats = (None, None, None)
        if not script.scaled_miller_array.space_group().is_centric():
            anom_stats = script.anom_merging_statistics_result
        logger.info(make_merging_statistics_summary(stats))
        try:
            if script.params.cut_data.d_min is None:
                d_min = resolution_cc_half(stats, limit=0.3).d_min
            else:
                d_min = script.params.cut_data.d_min
        except RuntimeError as e:
            logger.debug(f"Resolution fit failed: {e}")
        else:
            max_current_res = stats.bins[-1].d_min
            if d_min and d_min - max_current_res > 0.005:
                logger.info(
                    "Resolution limit suggested from CC"
                    + "\u00BD"
                    + " fit (limit CC"
                    + "\u00BD"
                    + "=0.3): %.2f",
                    d_min,
                )
                try:
                    cut_stats, cut_anom_stats = merging_stats_from_scaled_array(
                        script.scaled_miller_array.resolution_filter(d_min=d_min),
                        script.params.output.merging.nbins,
                        script.params.output.use_internal_variance,
                    )
                except DialsMergingStatisticsError:
                    pass
                else:
                    if script.scaled_miller_array.space_group().is_centric():
                        cut_anom_stats = None
        logger.info(table_1_summary(stats, anom_stats, cut_stats, cut_anom_stats))


class ScalingHTMLContextManager:
    def __init__(self, script):
        self.script = script

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        _make_scaling_html(self.script)


def _make_scaling_html(scaling_script):
    """Collect data from the individual observers and write the html."""
    html_file = scaling_script.params.output.html
    json_file = scaling_script.params.output.json
    if not (html_file or json_file):
        return
    data = {}
    data.update(
        make_scaling_model_plots(scaling_script.experiments, scaling_script.reflections)
    )
    data.update(
        make_outlier_plots(scaling_script.reflections, scaling_script.experiments)
    )
    data.update(
        make_error_model_plots(scaling_script.params, scaling_script.experiments)
    )
    data.update(make_merging_stats_plots(scaling_script))
    data.update(make_filtering_plots(scaling_script))
    if html_file:
        logger.info("Writing html report to %s", html_file)
        loader = ChoiceLoader(
            [
                PackageLoader("dials", "templates"),
                PackageLoader("dials", "static", encoding="utf-8"),
            ]
        )
        env = Environment(loader=loader)
        template = env.get_template("scaling_report.html")
        assert_is_json_serialisable(data, "self.data")
        html = template.render(
            page_title="DIALS scaling report",
            scaling_model_graphs=data["scaling_model"],
            scaling_tables=data["scaling_tables"],
            error_model_summary=data["error_model_summary"],
            resolution_plots=data["resolution_plots"],
            scaling_outlier_graphs=data["outlier_plots"],
            error_model_plots=data["error_model_plots"],
            anom_plots=data["anom_plots"],
            batch_plots=data["batch_plots"],
            image_range_tables=data["image_range_tables"],
            misc_plots=data["misc_plots"],
            filter_plots=data["filter_plots"],
        )
        with open(html_file, "wb") as f:
            f.write(html.encode("utf-8", "xmlcharrefreplace"))
    if json_file:
        logger.info("Writing html report data to %s", json_file)
        with open(json_file, "w", encoding="utf-8") as outfile:
            json.dump(data, outfile)


def make_scaling_model_plots(experiments, reflection_tables):
    """Collect scaling model plots for html report."""
    data = {i: e.scaling_model for i, e in enumerate(experiments)}
    d = {}
    combined_plots = make_combined_plots(data)
    if combined_plots:
        d.update(combined_plots)
    for key in sorted(data.keys()):
        scaling_model_plots = plot_scaling_models(data[key], reflection_tables[key])
        for plot in scaling_model_plots.values():
            plot["layout"]["title"] += f" (dataset {key})"
        for name, plot in scaling_model_plots.items():
            d[name + "_" + str(key)] = plot
    graphs = {"scaling_model": d}
    return graphs


def print_scaling_model_error_summary(experiments):
    """Get a summary of the error distribution of the models."""
    models = [e.scaling_model.to_dict() for e in experiments]
    first_model = models[0]
    component = first_model["configuration_parameters"]["corrections"][0]
    msg = ""
    if "est_standard_devs" in first_model[component]:
        p_sigmas = flex.double()
        for model in models:
            for component in model["configuration_parameters"]["corrections"]:
                if "est_standard_devs" in model[component]:
                    params = flex.double(model[component]["parameters"])
                    sigmas = flex.double(model[component]["est_standard_devs"])
                    null_value = flex.double(
                        len(params), model[component]["null_parameter_value"]
                    )
                    p_sigmas.extend(flex.abs(params - null_value) / sigmas)
        log_p_sigmas = flex.log(p_sigmas)
        frac_high_uncertainty = (log_p_sigmas < math.log(2)).count(True) / len(
            log_p_sigmas
        )
        if frac_high_uncertainty > 0.5:
            msg = (
                "Warning: Over half ({:.2f}%) of model parameters have significant\n"
                "uncertainty (sigma/abs(parameter) > 0.5), which could indicate a\n"
                "poorly-determined scaling problem or overparameterisation.\n"
            ).format(frac_high_uncertainty * 100)
        else:
            msg = (
                "{:.2f}% of model parameters have significant uncertainty\n"
                "(sigma/abs(parameter) > 0.5)\n"
            ).format(frac_high_uncertainty * 100)
    return msg


def make_outlier_plots(reflection_tables, experiments):
    """Make outlier plots for the HTML report."""
    data = {}
    for j, (table, expt) in enumerate(zip(reflection_tables, experiments)):
        outliers = table.get_flags(table.flags.outlier_in_scaling)
        x, y, z = table["xyzobs.px.value"].select(outliers).parts()
        if expt.scan:
            zrange = [
                i / expt.scan.get_oscillation()[1]
                for i in expt.scan.get_oscillation_range()
            ]
        else:
            zrange = [0, 0]
        data[j] = {
            "x": list(x),
            "y": list(y),
            "z": list(z),
            "image_size": expt.detector[0].get_image_size(),
            "z_range": zrange,
        }
    d = {}
    for key in sorted(data):
        outlier_plots = plot_outliers(data[key])
        for plot in outlier_plots.values():
            if plot:  # may be null if no outliers
                plot["layout"]["title"] += f" (dataset {key})"
        d["outlier_plot_" + str(key)] = outlier_plots["outlier_xy_positions"]
        d["outlier_plot_z" + str(key)] = outlier_plots["outliers_vs_z"]
    graphs = {"outlier_plots": d}
    return graphs


def make_error_model_plots(params, experiments):
    """Generate normal probability plot data."""
    d = {"error_model_plots": {}, "error_model_summary": "No error model applied"}
    error_model_data = []  # a list of dicts of error model data
    if experiments[0].scaling_model.error_model:
        error_models = [e.scaling_model.error_model for e in experiments]
        unique_error_models = OrderedSet(error_models)
        if len(unique_error_models) == 1:
            d["error_model_summary"] = str(error_models[0])
        else:
            d["error_model_summary"] = ""
            for i, e in enumerate(unique_error_models):
                indices = [str(j + 1) for j, x in enumerate(error_models) if e is x]
                d["error_model_summary"] += (
                    f"\nError model {i+1}, applied to sweeps {', '.join(indices)}:"
                    + str(e)
                )
        for em in unique_error_models:
            if em.filtered_Ih_table:
                data_i = {}
                table = em.filtered_Ih_table
                data_i["intensity"] = table.intensities
                sigmaprime = calc_sigmaprime(em.parameters, table)
                data_i["delta_hl"] = calc_deltahl(table, table.calc_nh(), sigmaprime)
                data_i["inv_scale"] = table.inverse_scale_factors
                data_i["sigma"] = sigmaprime * data_i["inv_scale"]
                data_i["binning_info"] = em.binner.binning_info
                em.clear_Ih_table()
                if params.weighting.error_model.basic.minimisation == "regression":
                    x, y = calculate_regression_x_y(em.filtered_Ih_table)
                    data_i["regression_x"] = x
                    data_i["regression_y"] = y
                    data_i["model_a"] = em.parameters[0]
                    data_i["model_b"] = em.parameters[1]
                error_model_data.append(data_i)

    if error_model_data:
        for i, emd in enumerate(error_model_data):
            d["error_model_plots"].update(normal_probability_plot(emd, label=i + 1))
            d["error_model_plots"].update(
                i_over_sig_i_vs_i_plot(
                    flumpy.from_numpy(emd["intensity"]),
                    flumpy.from_numpy(emd["sigma"]),
                    label=i + 1,
                )
            )
            d["error_model_plots"].update(error_model_variance_plot(emd, label=i + 1))
            if "regression_x" in emd:
                d["error_model_plots"].update(error_regression_plot(emd, label=i + 1))
    return d


def make_filtering_plots(script):
    """Make filtering plots for HTML report"""
    if script.filtering_results:
        data = {
            "merging_stats": script.filtering_results.get_merging_stats(),
            "initial_expids_and_image_ranges": script.filtering_results.initial_expids_and_image_ranges,
            "cycle_results": script.filtering_results.get_cycle_results(),
            "expids_and_image_ranges": script.filtering_results.expids_and_image_ranges,
            "mode": script.params.filtering.deltacchalf.mode,
        }
        d = make_scaling_filtering_plots(data)
        return {"filter_plots": d}
    return {"filter_plots": {}}


def make_merging_stats_plots(script, run_xtriage_analysis=False, make_batch_plots=True):
    """Make merging stats plots for HTML report"""
    d = {
        "scaling_tables": {},
        "resolution_plots": {},
        "batch_plots": {},
        "misc_plots": {},
        "anom_plots": {},
        "image_range_tables": [],
    }
    if make_batch_plots:
        (
            batches,
            rvb,
            isigivb,
            svb,
            batch_data,
        ) = reflection_tables_to_batch_dependent_properties(  # pylint: disable=unbalanced-tuple-unpacking
            script.reflections,
            script.experiments,
            script.scaled_miller_array,
        )
        bm = batch_manager(batches, batch_data)
        image_range_tables = make_image_range_table(script.experiments, bm)
        d["image_range_tables"] = [image_range_tables]
        d["batch_plots"].update(scale_rmerge_vs_batch_plot(bm, rvb, svb))
        d["batch_plots"].update(i_over_sig_i_vs_batch_plot(bm, isigivb))

    if script.merging_statistics_result:
        stats = script.merging_statistics_result
        anom_stats = script.anom_merging_statistics_result
        is_centric = script.scaled_miller_array.space_group().is_centric()
        # Now calculate batch data
        plotter = ResolutionPlotsAndStats(stats, anom_stats, is_centric)
        d["resolution_plots"].update(plotter.make_all_plots())
        d["scaling_tables"] = plotter.statistics_tables()
        plotter = IntensityStatisticsPlots(
            script.scaled_miller_array, run_xtriage_analysis=run_xtriage_analysis
        )
        d["xtriage_output"] = plotter.generate_xtriage_output()
        d["resolution_plots"].update(plotter.generate_resolution_dependent_plots())
        if d["resolution_plots"]["cc_one_half"]["data"][2]:
            cc_anom = d["resolution_plots"]["cc_one_half"]["data"][2]["y"]
            significance = d["resolution_plots"]["cc_one_half"]["data"][3]["y"]
            sig = flex.double(cc_anom) > flex.double(significance)
            max_anom = 0
            for i, v in enumerate(sig):
                if v:
                    max_anom = i
                else:
                    break
            d_min = uctbx.d_star_sq_as_d(plotter.binner.limits())[max_anom + 1]
        else:
            d_min = 0.0
        d["misc_plots"].update(plotter.generate_miscellanous_plots())
        intensities_anom = script.scaled_miller_array.as_anomalous_array()
        intensities_anom = intensities_anom.map_to_asu().customized_copy(
            info=script.scaled_miller_array.info()
        )
        anom_plotter = AnomalousPlotter(intensities_anom, strong_cutoff=d_min)
        d["anom_plots"].update(anom_plotter.make_plots())
    return d



from __future__ import annotations

from iotbx import cif, mtz, pdb
from mmtbx.command_line.fmodel import fmodel_from_xray_structure_master_params
from mmtbx.utils import fmodel_from_xray_structure


def intensities_from_reference_file(filename, d_min=2.0, wavelength=None):
    """
    Extract/calculate intensities from a reference file - which may contain data
    or a model, from a pdb, cif or mtz file. For cif, this can be a file
    containing measured data, or a pdb model conforming to mmcif, or an inorgranic
    cif conforming to core_cif.
    """
    if filename.endswith(".pdb"):
        return intensity_array_from_pdb_model_file(filename, d_min, wavelength)
    elif filename.endswith(".cif"):
        # need to see if file is a datafile or model file.
        # First try to interpret as a data file (quick and cheap to try)
        try:
            return intensity_array_from_cif_data_file(filename)
        except KeyError:
            return intensity_array_from_cif_model_file(filename, d_min, wavelength)
    elif filename.endswith(".mtz"):
        return intensity_array_from_mtz_file(filename)
    else:
        raise ValueError(
            "Unrecognised input format for reference file (expected .pdb, .cif or .mtz)"
        )


def intensities_from_reference_data_file(filename):
    """
    Extract intensities from a reference data file (i.e. cif or mtz containing
    intensities/measured F values.
    """
    if filename.endswith(".cif"):
        return intensity_array_from_cif_data_file(filename)
    elif filename.endswith(".mtz"):
        return intensity_array_from_mtz_file(filename)
    else:
        raise ValueError(
            "Unrecognised input format for reference data file (expected .cif or .mtz)"
        )


def intensities_from_reference_model_file(filename, d_min=2.0, wavelength=None):
    """
    Calculate intensities from a reference model file (i.e. cif or pdb containing
    a crystal structure).
    """
    if filename.endswith(".cif"):
        return intensity_array_from_cif_model_file(filename, d_min, wavelength)
    elif filename.endswith(".pdb"):
        return intensity_array_from_pdb_model_file(filename, d_min, wavelength)
    else:
        raise ValueError(
            "Unrecognised input format for reference model file (expected .cif or .pdb)"
        )


def intensity_array_from_cif_model_file(cif_file, d_min=2.0, wavelength=None):
    """Return an intensity miller array from a cif file."""
    try:
        # First try to interpret as a MX cif data structure
        xray_structure = pdb.hierarchy.input(cif_file).xray_structure_simple()
    except AssertionError:
        # If this is not understood, try to read as an inorganic cif, using
        # direct methods to calculate the intensities.
        r = cif.reader(file_path=cif_file)
        xray_structure = list(r.build_crystal_structures().values())[0]
        ic = (
            xray_structure.structure_factors(
                anomalous_flag=True, d_min=d_min, algorithm="direct"
            )
            .f_calc()
            .as_intensity_array()
        )
    else:
        # Use the mmtbx methods for quick calculation of fs from model.
        ic = _mmtbx_intensity_from_structure(xray_structure, d_min, wavelength)
    return ic


def intensity_array_from_pdb_model_file(pdb_file, d_min=2.0, wavelength=None):
    xray_structure = pdb.hierarchy.input(pdb_file).xray_structure_simple()
    ic = _mmtbx_intensity_from_structure(xray_structure, d_min, wavelength)
    return ic


def _mmtbx_intensity_from_structure(xray_structure, d_min, wavelength=None):
    """Use the mmtbx methods for quick calculation of fs from model."""
    if wavelength:
        xray_structure.set_inelastic_form_factors(photon=wavelength, table="sasaki")
    params = fmodel_from_xray_structure_master_params.extract()
    params.high_resolution = d_min
    if wavelength:
        params.wavelength = wavelength
    fm = fmodel_from_xray_structure(xray_structure, params=params)
    ic = fm.f_model.as_intensity_array()
    return ic


def intensity_array_from_cif_data_file(cif_file):
    f = cif.reader(file_path=cif_file)
    # N.B. in mmcif files, the item is _refln.F_meas (or _refln.F_meas_au),
    # whereas for 'core_cif' inorganic cif the item is _refln_F_meas. This will
    # pick up both cases.
    for ma in f.as_miller_arrays():
        for l in ma.info().labels:
            if "F_meas" in l:
                return ma.as_intensity_array()
    raise KeyError("Unable to find F_meas in cif file")


def intensity_array_from_mtz_file(filename):
    m = mtz.object(filename)
    col_keys = [c.label() for c in m.columns()]
    # try to make an anomalous intensity array first
    anom_labels = ["I(+)", "SIGI(+)", "I(-)", "SIGI(-)"]
    normal_labels = ["I", "SIGI"]
    mean_labels = ["IMEAN", "SIGIMEAN"]
    F_anom_labels = ["F(+)", "SIGF(+)", "F(-)", "SIGF(-)"]
    F_normal_labels = ["F", "SIGF"]
    I_labels_to_try = [anom_labels, normal_labels, mean_labels]
    F_labels_to_try = [F_anom_labels, F_normal_labels]
    for labels in I_labels_to_try:
        if all(k in col_keys for k in labels):
            for ma in m.as_miller_arrays():
                if ma.info().labels == labels:
                    return ma
    for labels in F_labels_to_try:
        if all(k in col_keys for k in labels):
            for ma in m.as_miller_arrays():
                if ma.info().labels == labels:
                    return ma.as_intensity_array()
    else:
        raise KeyError("Unable to extract intensity array from mtz file")



from __future__ import annotations

import logging
import math
import sys

import numpy as np

from scitbx.matrix import col

from dials.algorithms.integration.kapton_correction import get_absorption_correction
from dials.algorithms.shoebox import MaskCode
from dials.array_family import flex

logging.basicConfig()
logger = logging.getLogger(__name__)


class KaptonTape_2019:
    """Class for defining Kapton tape using dxtbx models and finding the path through the tape traversed by s1 vector"""

    def __init__(
        self,
        height_mm,
        thickness_mm,
        half_width_mm,
        rotation_angle_deg,
        wavelength_ang=None,
    ):
        self.height_mm = h = height_mm  # plugin controlled
        self.thickness_mm = t = thickness_mm  # plugin controlled
        self.half_width_mm = w = half_width_mm  # plugin controlled
        self.tape_depth_mm = a = half_width_mm * 20.0  # Big number in mm
        self.angle_rad = rotation_angle_deg * math.pi / 180.0  # plugin controlled
        self.wavelength_ang = wavelength_ang
        self.num_pixels = 5000  # number of pixels to put in fictitious kapton faces
        # Now set up the kapton physical model using dxtbx detector objects
        #
        # determine absorption coeff (mm-1) through kapton for a given X-ray energy
        G = get_absorption_correction()
        attenuation_length_mm = G(self.wavelength_ang)
        self.abs_coeff = 1 / attenuation_length_mm

        def create_kapton_face(ori, fast, slow, image_size, pixel_size, name):
            """Create a face of the kapton as a dxtbx detector object"""
            from dxtbx.model import Detector

            d = Detector()
            p = d.add_panel()
            p.set_local_frame(fast.elems, slow.elems, ori.elems)
            p.set_pixel_size((pixel_size, pixel_size))
            p.set_image_size(image_size)
            p.set_trusted_range((0, 2e6))
            p.set_name(f"KAPTON_{name}")
            return d

        # Set up the bounding box of the kapton
        rA = col((t / 2, a / 2, -w))
        rB = col((t / 2, a / 2, w))
        rC = col((t / 2, -a / 2, w))
        rD = col((t / 2, -a / 2, -w))
        rE = col((-t / 2, a / 2, -w))
        rF = col((-t / 2, a / 2, w))
        rG = col((-t / 2, -a / 2, w))
        rH = col((-t / 2, -a / 2, -w))
        # Now add a rotation
        # rotation about z-axis
        rot_axis = col((0, 0, 1))
        rot_mat = rot_axis.axis_and_angle_as_r3_rotation_matrix(
            self.angle_rad, deg=False
        )

        rA = rot_mat * rA
        rB = rot_mat * rB
        rC = rot_mat * rC
        rD = rot_mat * rD
        rE = rot_mat * rE
        rF = rot_mat * rF
        rG = rot_mat * rG
        rH = rot_mat * rH
        # Now add an offset to all the points in the direction normal to the ABCD plane
        rAB = rA - rB
        rCB = rC - rB
        # Get normal in the direction of the drop
        drop_normal = rCB.cross(rAB).normalize()
        offset = -(h + t / 2) * drop_normal
        rA = rA + offset
        rB = rB + offset
        rC = rC + offset
        rD = rD + offset
        rE = rE + offset
        rF = rF + offset
        rG = rG + offset
        rH = rH + offset
        # Store these edge points if needed later
        self.edge_points = [rA, rB, rC, rD, rE, rF, rG, rH]
        # Now set up the 6 faces
        faces = []
        px_off = 0
        fast = (rA - rB).normalize()
        slow = (rC - rB).normalize()
        pixel_size = self.half_width_mm * 2.0 / self.num_pixels
        image_size = (
            self.num_pixels + px_off,
            int(self.tape_depth_mm / pixel_size) + px_off,
        )
        # Face YZ0
        ori = rB
        faces.append(create_kapton_face(ori, fast, slow, image_size, pixel_size, "yz0"))
        # face YZ1
        ori = rF
        faces.append(create_kapton_face(ori, fast, slow, image_size, pixel_size, "yz1"))
        #
        fast = (rB - rF).normalize()
        slow = (rG - rF).normalize()
        pixel_size = self.thickness_mm / self.num_pixels
        image_size = (
            self.num_pixels + px_off,
            int(self.tape_depth_mm / pixel_size) + px_off,
        )
        # Face XY0
        ori = rF
        faces.append(create_kapton_face(ori, fast, slow, image_size, pixel_size, "xy0"))
        # face XY1
        ori = rE
        faces.append(create_kapton_face(ori, fast, slow, image_size, pixel_size, "xy1"))
        #
        fast = (rA - rE).normalize()
        slow = (rF - rE).normalize()
        pixel_size = self.thickness_mm / self.num_pixels
        image_size = (
            self.num_pixels + px_off,
            int(self.half_width_mm * 2.0 / pixel_size) + px_off,
        )
        # Face XZ0
        ori = rE
        faces.append(create_kapton_face(ori, fast, slow, image_size, pixel_size, "xz0"))
        # face XY1
        ori = rH
        faces.append(create_kapton_face(ori, fast, slow, image_size, pixel_size, "xz1"))
        #
        self.faces = faces

    def get_kapton_path_mm(self, s1):
        """Get kapton path length traversed by an s1 vecto. If no kapton intersection or just touches the edge,
        then None is returned"""
        intersection_points = []
        # determine path length through kapton tape
        for face in self.faces:
            try:
                px = [int(p) for p in face[0].get_ray_intersection_px(s1)]
                # faces are single panel anyway
                if (
                    px[0] < 0
                    or px[1] < 0
                    or px[0] > face[0].get_image_size()[0]
                    or px[1] > face[0].get_image_size()[1]
                ):
                    continue
                intersection_points.append(
                    face[0].get_lab_coord(face[0].get_ray_intersection(s1))
                )  # faces are single panel anyway
            except RuntimeError:
                pass

        if not intersection_points:
            return 0.0
        if len(intersection_points) == 1:
            logger.warning(
                "Ray not intersecting 2 faces.Either s1 vector does not intersect kapton or just touches edge of kapton tape \
                    No correction is needed for those cases."
            )
            return 0.0

        n_intersection_points = len(intersection_points)
        kapton_path_mm = []
        for ii in range(n_intersection_points - 1):
            for jj in range(ii + 1, n_intersection_points):
                kapton_path_mm.append(
                    (
                        col(intersection_points[ii]) - col(intersection_points[jj])
                    ).length()
                )
        return max(kapton_path_mm)

    def abs_correction(self, s1):
        """Compute absorption correction using beers law. Takes in a tuple for a single s1 vector and does absorption correction"""
        kapton_path_mm = self.get_kapton_path_mm(s1)
        if kapton_path_mm is not None:
            absorption_correction = 1 / math.exp(
                -self.abs_coeff * kapton_path_mm
            )  # unitless, >=1
        return absorption_correction

    def abs_correction_flex(self, s1_flex):
        """Compute the absorption correction using beers law. Takes in a flex array of s1 vectors, determines path lengths for each
        and then determines absorption correction for each s1 vector"""
        kapton_faces = self.faces
        from dials.algorithms.integration import get_kapton_path_cpp

        # new style, much faster
        # Note, the last two faces should never be hit by a photon so don't need to check them
        kapton_path_mm = get_kapton_path_cpp(kapton_faces[:4], s1_flex)
        # old style, really slow
        # for s1 in s1_flex:
        #  kapton_path_mm.append(self.get_kapton_path_mm(s1))
        # determine absorption correction
        if kapton_path_mm is not None:
            absorption_correction = 1 / flex.exp(
                -self.abs_coeff * kapton_path_mm
            )  # unitless, >=1
        return absorption_correction

    def distance_of_point_from_line(self, r0, r1, r2):
        """Evaluates distance between point and a line between two points
        Note that implementation ignores z dimension"""
        x0, y0, z0 = r0
        x1, y1, z1 = r1
        x2, y2, z2 = r2
        num = (y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1
        denom = math.sqrt((y2 - y1) * (y2 - y1) + (x2 - x1) * (x2 - x1))
        return abs(num) / denom

    def get_edge_distances(self, edges):
        dist_list = flex.double()
        dist_list_idx = []
        n_edges = len(edges)
        for edge_1 in range(n_edges - 1):
            pt1 = col(edges[edge_1])
            for edge_2 in range(edge_1 + 1, n_edges):
                pt2 = col(edges[edge_2])
                distance = (pt1 - pt2).length()
                dist_list.append(distance)
                dist_list_idx.append((edge_1, edge_2))

        return dist_list, dist_list_idx

    def abs_bounding_lines_in_mm(self, detector):
        """Return bounding lines of kapton"""

        def _check_int_edge_pts(int_edge_pts):
            """Function to ensure that the combination of int_edge_pts will not result
            in a kapton edge to be defined by 2 identical int_edge_pts.
            """
            new_int_edge_pts = int_edge_pts.copy()
            if len(set(int_edge_pts)) == 4:
                return int_edge_pts
            elif len(set(int_edge_pts)) == 2:
                sys.exit(
                    "Insuffient number of intersection points to define both Kapton edges"
                )
            else:
                # Find different permuations of intersecting kapton edge points that won't
                # result in kapton_edges defined by identical points
                if (
                    int_edge_pts[0] == int_edge_pts[1]
                    or int_edge_pts[2] == int_edge_pts[3]
                ):
                    new_int_edge_pts[1] = int_edge_pts[3]
                    new_int_edge_pts[3] = int_edge_pts[1]
                return new_int_edge_pts

        def _orient_det_edge_pts(det_edge_pts):
            """
            Function to orient the detector edge points counterclockwise
            such that the first edge point index is in the upper left quadrant
            (i.e., -x, +y). This isn't explictly necessary but nevertheless useful
            b/c the function abs_bounding_lines_on_image always assumes that the
            detector edges used to help calculate the Kapton absorption max and edge
            are a parallel to the y-axis. This function enforces that a convention
            so that this assumption in the later logic will be true.
            """
            assert len(det_edge_pts) == 4, "Detectors can only be defined by 4 points"
            n_edges = float(len(det_edge_pts))
            # find center of detector
            x = [p[0] for p in det_edge_pts]
            y = [p[1] for p in det_edge_pts]
            z = [p[2] for p in det_edge_pts]
            center = [sum(x) / n_edges, sum(y) / n_edges, sum(z) / n_edges]
            # define reference vectors used to determine rel. orientation
            x_ref = [center[0] + 100, center[1], center[2]]
            y_ref = [center[0], center[1] + 100, center[2]]
            x_ref_vec = (col(x_ref) - col(center)).normalize()
            y_ref_vec = (col(y_ref) - col(center)).normalize()
            # determine orientation of each edge pt wrt x and y axis
            x_axis_angles = []
            y_axis_angles = []
            for edge in det_edge_pts:
                vec = (col(edge) - col(center)).normalize()
                x_axis_angles.append(np.degrees(np.arccos(vec.dot(x_ref_vec))))
                y_axis_angles.append(np.degrees(np.arccos(vec.dot(y_ref_vec))))
            x_axis_angles = [
                (90.0 - xang) if xang > 90.0 else xang for xang in x_axis_angles
            ]
            y_axis_angles = [
                (90.0 - yang) if yang > 90.0 else yang for yang in y_axis_angles
            ]
            # assign edge pt indices based on quadrant position starting in 4th quadrant (-x.-y)
            # proceeding counterclockwise
            edge_0 = np.where(
                (np.asarray(x_axis_angles) < 0) & (np.asarray(y_axis_angles) > 0)
            )[0][0]
            edge_1 = np.where(
                (np.asarray(x_axis_angles) < 0) & (np.asarray(y_axis_angles) < 0)
            )[0][0]
            edge_2 = np.where(
                (np.asarray(x_axis_angles) > 0) & (np.asarray(y_axis_angles) < 0)
            )[0][0]
            edge_3 = np.where(
                (np.asarray(x_axis_angles) > 0) & (np.asarray(y_axis_angles) > 0)
            )[0][0]

            return [
                det_edge_pts[edge_0],
                det_edge_pts[edge_1],
                det_edge_pts[edge_2],
                det_edge_pts[edge_3],
            ]

        def _make_kapton_distance_vec(
            det_edge_pt1, det_edge_pt2, int_edge_pt1, int_edge_pt2
        ):
            """
            Function to create vectors that descibe the distance btw two points where rays from
            Kapton edges intersect with detector relative to detector edges
            """
            dist_det_edge_x = det_edge_pt2[0] - det_edge_pt1[0]
            dist_det_edge_y = det_edge_pt2[1] - det_edge_pt1[1]
            mid_pt = [
                det_edge_pt1[0] + dist_det_edge_x / 2,
                det_edge_pt1[1] + dist_det_edge_y / 2,
            ]
            kapton_int_pt1_vec = (
                col([int_edge_pt1[0], int_edge_pt1[1]]) - col([mid_pt[0], mid_pt[1]])
            ).normalize()
            kapton_int_pt2_vec = (
                col([int_edge_pt2[0], int_edge_pt2[1]]) - col([mid_pt[0], mid_pt[1]])
            ).normalize()

            return kapton_int_pt1_vec, kapton_int_pt2_vec

        # first get bounding directions from detector:
        detz = flex.mean(flex.double([panel.get_origin()[2] for panel in detector]))
        edges = []
        for ii, panel in enumerate(detector):
            f_size, s_size = panel.get_image_size()
            for point in [(0, 0), (0, s_size), (f_size, 0), (f_size, s_size)]:
                x, y = panel.get_pixel_lab_coord(point)[0:2]
                edges.append((x, y, detz))
        # Use the idea that the corners of the detector are end points of the diagonal and
        # will be top 2 max dimension among all end points
        dlist, dlist_idx = self.get_edge_distances(edges)
        sorted_idx = flex.sort_permutation(dlist, reverse=True)

        edge_pts = [
            edges[dlist_idx[sorted_idx[0]][0]],
            edges[dlist_idx[sorted_idx[1]][0]],
            edges[dlist_idx[sorted_idx[0]][1]],
            edges[dlist_idx[sorted_idx[1]][1]],
        ]

        self.detector_edges = edge_pts
        edge_pts = _orient_det_edge_pts(edge_pts)
        # Now get the maximum extent of the intersection of the rays with the detector
        all_ints = []
        kapton_path_list = []

        for ii, edge_point in enumerate(self.edge_points):
            s1 = edge_point.normalize()
            kapton_path_mm = self.get_kapton_path_mm(s1)
            for panel in detector:
                try:
                    x_int, y_int = panel.get_lab_coord(panel.get_ray_intersection(s1))[
                        0:2
                    ]
                except RuntimeError:
                    pass
                int_point = (x_int, y_int, detz)
                # Arbitrary tolerance of couple of pixels otherwise these points
                # were getting clustered together
                tolerance = min(panel.get_pixel_size()) * 2.0
                if (
                    sum(
                        (col(trial_pt) - col(int_point)).length() <= tolerance
                        for trial_pt in all_ints
                    )
                    == 0
                ):
                    all_ints.append(int_point)
                kapton_path_list.append(kapton_path_mm)
        # Use the idea that the extreme edges of the intersection points are end points
        # of the diagonal and will be the top 2 max dimension among all end points
        dlist, dlist_idx = self.get_edge_distances(all_ints)
        sorted_idx = flex.sort_permutation(dlist, reverse=True)

        int_edge_pts = [
            all_ints[dlist_idx[sorted_idx[0]][0]],
            all_ints[dlist_idx[sorted_idx[1]][0]],
            all_ints[dlist_idx[sorted_idx[0]][1]],
            all_ints[dlist_idx[sorted_idx[1]][1]],
        ]
        int_edge_pts = _check_int_edge_pts(int_edge_pts)

        # Sort out the edge points and the int_edge_points which are on the same side
        kapton_edge_1 = (col(int_edge_pts[0]) - col(int_edge_pts[1])).normalize()
        kapton_edge_2 = (col(int_edge_pts[2]) - col(int_edge_pts[3])).normalize()
        # Make sure the edges of the detector and the kapton are in the same orientation
        # first for kapton edge 1
        edge_idx = (0, 1, 2, 3)
        side_1 = (col(edge_pts[edge_idx[0]]) - col(edge_pts[edge_idx[1]])).normalize()
        side_2 = (col(edge_pts[edge_idx[2]]) - col(edge_pts[edge_idx[3]])).normalize()
        v1 = kapton_edge_1.dot(side_1)
        v2 = kapton_edge_2.dot(side_2)
        if v1 < 0.0:
            edge_idx = (edge_idx[1], edge_idx[0], edge_idx[2], edge_idx[3])
        if v2 < 0.0:
            edge_idx = (edge_idx[0], edge_idx[1], edge_idx[3], edge_idx[2])

        # Define two vectors that describe the positions of the top two intersecion edge pts
        # (intersection of the kapton rays with the detector) relative to the center of the
        # detector. These vectors will allow us to determine and assign which intersection
        # edge points are associated with the absorption max or edge (aka min) edge.
        kapton_int_pt1_vector, kapton_int_pt2_vector = _make_kapton_distance_vec(
            edge_pts[edge_idx[0]],
            edge_pts[edge_idx[3]],
            int_edge_pts[0],
            int_edge_pts[3],
        )
        if kapton_int_pt1_vector[0] == kapton_int_pt2_vector[0]:
            # the same kapton int_pt was used to define a kapton edge twice
            # try another one
            kapton_int_pt1_vector, kapton_int_pt2_vector = _make_kapton_distance_vec(
                edge_pts[edge_idx[0]],
                edge_pts[edge_idx[3]],
                int_edge_pts[1],
                int_edge_pts[2],
            )
        int_edge_idx = None
        # abs max edge is left of min edge
        if (kapton_int_pt1_vector[0] < 0) & (kapton_int_pt2_vector[0] > 0):
            int_edge_idx = [0, 1, 2, 3]
        # both edges are on the left side of det
        elif (kapton_int_pt1_vector[0] < 0) & (kapton_int_pt2_vector[0] < 0):
            # abs max edge to the left of min edge
            if kapton_int_pt1_vector[0] < kapton_int_pt2_vector[0]:
                int_edge_idx = [2, 3, 0, 1]
            # abs max edge to the right of min edge
            elif kapton_int_pt1_vector[0] > kapton_int_pt2_vector[0]:
                int_edge_idx = [0, 1, 2, 3]
        # both edges are on the right side of det
        elif (kapton_int_pt1_vector[0] > 0) & (kapton_int_pt2_vector[0] > 0):
            # abs max edge is to right of min edge
            if kapton_int_pt1_vector[0] > kapton_int_pt2_vector[0]:
                int_edge_idx = [2, 3, 0, 1]
            # abs max edge is the the left of min edge
            elif kapton_int_pt1_vector[0] < kapton_int_pt2_vector[0]:
                int_edge_idx = [0, 1, 2, 3]
        # abs max edge is right of min edge
        elif (kapton_int_pt1_vector[0] > 0) & (kapton_int_pt2_vector[0] < 0):
            int_edge_idx = [2, 3, 0, 1]

        pair_values = [
            (
                edge_pts[edge_idx[0]],
                edge_pts[edge_idx[1]],
                edge_pts[edge_idx[2]],
                edge_pts[edge_idx[3]],
            ),
            (
                int_edge_pts[int_edge_idx[0]],
                int_edge_pts[int_edge_idx[1]],
                int_edge_pts[int_edge_idx[2]],
                int_edge_pts[int_edge_idx[3]],
            ),
        ]

        return pair_values

    def abs_bounding_lines_on_image(self, detector):
        pair_values = self.abs_bounding_lines_in_mm(detector)
        r0, r1, r2, r3 = pair_values[0]
        ra, rb, rc, rd = pair_values[1]
        # Get slope, intercept from 0-2 edge line and 1-3 edge line

        def get_line_equation(rA, rB):
            xA, yA, zA = rA
            xB, yB, zB = rB
            m = (yB - yA) / (xB - xA)
            c = yA - m * xA
            return (m, c)

        def get_line_intersection(m1, c1, m2, c2):
            x = (c2 - c1) / (m1 - m2)
            y = m1 * x + c1
            return (x, y)

        m03, c03 = get_line_equation(r0, r3)
        m12, c12 = get_line_equation(r1, r2)

        # Get slope, intercept for a-b and c-d
        mab, cab = get_line_equation(ra, rb)
        mcd, ccd = get_line_equation(rc, rd)
        # Get intersection between 0-2 with a-b
        r03_ab = get_line_intersection(m03, c03, mab, cab)
        # Get line intersection between 1-3 and a-b
        r12_ab = get_line_intersection(m12, c12, mab, cab)
        # Get line intersection between 0-2 and c-d
        r03_cd = get_line_intersection(m03, c03, mcd, ccd)
        # Get line intersection between 1-3 and c-d
        r12_cd = get_line_intersection(m12, c12, mcd, ccd)
        # returns pairs of x,y
        return [
            (r03_ab[0], r03_ab[1], r12_ab[0], r12_ab[1]),
            (r03_cd[0], r03_cd[1], r12_cd[0], r12_cd[1]),
        ]


class image_kapton_correction:
    def __init__(
        self,
        panel_size_px=None,  #
        pixel_size_mm=None,  #
        detector_dist_mm=None,  #
        wavelength_ang=None,
        reflections_sele=None,
        params=None,
        expt=None,
        refl=None,
        smart_sigmas=True,
        logger=None,
    ):
        self.panel_size_px = panel_size_px
        self.pixel_size_mm = pixel_size_mm
        self.detector_dist_mm = detector_dist_mm
        self.wavelength_ang = wavelength_ang
        self.reflections_sele = reflections_sele
        self.params = params
        self.expt = expt
        self.refl = refl
        self.smart_sigmas = smart_sigmas
        self.logger = logger
        self.extract_params()

    def extract_params(self):
        h = self.params.xtal_height_above_kapton_mm.value
        t = self.params.kapton_thickness_mm.value
        w = self.params.kapton_half_width_mm.value
        a = self.params.rotation_angle_deg.value
        self.kapton_params = (h, t, w, a)

        if self.smart_sigmas:
            sig_h = self.params.xtal_height_above_kapton_mm.sigma
            sig_t = self.params.kapton_thickness_mm.sigma
            sig_w = self.params.kapton_half_width_mm.sigma
            sig_a = self.params.rotation_angle_deg.sigma
            self.kapton_params_sigmas = (sig_h, sig_t, sig_w, sig_a)
            assert all(
                sig >= 0 for sig in self.kapton_params_sigmas
            ), "Kapton param sigmas must be non-negative"
            self.kapton_params_maxes = [
                [
                    self.kapton_params[i] + self.kapton_params_sigmas[j]
                    if j == i
                    else self.kapton_params[i]
                    for i in range(4)
                ]
                for j in range(4)
            ]
            self.kapton_params_mins = [
                [
                    max(self.kapton_params[i] - self.kapton_params_sigmas[j], 0.001)
                    if j == i
                    else self.kapton_params[i]
                    for i in range(3)
                ]
                + [a]
                for j in range(3)
            ] + [[self.kapton_params[i] for i in range(3)] + [a - sig_a]]

    def __call__(self, plot=False):
        def correction_and_within_spot_sigma(params_version, variance_within_spot=True):
            # instantiate Kapton absorption class here
            absorption = KaptonTape_2019(
                params_version[0],
                params_version[1],
                params_version[2],
                params_version[3],
                self.wavelength_ang,
            )
            # *map(float, self.panel_size_px)) #
            detector = self.expt.detector

            absorption_corrections = flex.double()
            absorption_sigmas = (
                flex.double()
            )  # std dev of corrections for pixels within a spot, default sigma

            if variance_within_spot:
                mask_code = MaskCode.Foreground | MaskCode.Valid
                for iref in range(len(self.reflections_sele)):
                    kapton_correction_vector = flex.double()
                    # foreground: integration mask
                    shoebox = self.reflections_sele[iref]["shoebox"]
                    foreground = (
                        (shoebox.mask.as_1d() & mask_code) == mask_code
                    ).iselection()
                    f_absolute, s_absolute, z_absolute = (
                        shoebox.coords().select(foreground).parts()
                    )
                    panel_number = self.reflections_sele[iref]["panel"]
                    lab_coords = detector[panel_number].get_lab_coord(
                        detector[panel_number].pixel_to_millimeter(
                            flex.vec2_double(f_absolute, s_absolute)
                        )
                    )
                    s1 = lab_coords.each_normalize()
                    # Real step right here
                    kapton_correction_vector.extend(absorption.abs_correction_flex(s1))
                    #
                    average_kapton_correction = flex.mean(kapton_correction_vector)
                    absorption_corrections.append(average_kapton_correction)
                    try:
                        spot_px_stddev = flex.mean_and_variance(
                            kapton_correction_vector
                        ).unweighted_sample_standard_deviation()
                    except Exception:
                        assert (
                            len(kapton_correction_vector) == 1
                        ), "stddev could not be calculated"
                        spot_px_stddev = 0
                    absorption_sigmas.append(spot_px_stddev)
                return absorption_corrections, absorption_sigmas
            else:
                s1_flex = self.reflections_sele["s1"].each_normalize()
                absorption_corrections = absorption.abs_correction_flex(s1_flex)
                return absorption_corrections, None

        # loop through modified Kapton parameters to get alternative corrections and estimate sigmas as
        # maximum variation between these versions of the corrections, on a per-spot basis, or the standard
        # deviation within a single spot, whichever is larger.
        self.logger.info("Calculating kapton corrections to integrated intensities...")
        corrections, sigmas = correction_and_within_spot_sigma(
            self.kapton_params, variance_within_spot=self.params.within_spot_sigmas
        )
        if self.smart_sigmas:
            for p in self.kapton_params_mins + self.kapton_params_maxes:
                self.logger.info("Calculating smart sigmas...")
                modif_corrections, _ = correction_and_within_spot_sigma(
                    p, variance_within_spot=False
                )
                perturbed = flex.abs(corrections - modif_corrections)
                if sigmas is None:
                    sigmas = perturbed
                else:
                    replace_sel = perturbed > sigmas
                    sigmas.set_selected(replace_sel, perturbed.select(replace_sel))
        if plot:
            from matplotlib import pyplot as plt

            for (title, data) in [("corrections", corrections), ("sigmas", sigmas)]:
                plt.hist(data, 20)
                plt.title(title)
                plt.show()
        if self.logger is not None:
            self.logger.info(
                "Returning absorption corrections and sigmas for %d spots"
                % len(corrections)
            )
        return corrections, sigmas


class multi_kapton_correction:
    def __init__(self, experiments, integrated, kapton_params, logger=None):
        self.experiments = experiments
        self.reflections = integrated
        self.params = kapton_params
        self.logger = logger

    def __call__(self):
        self.corrected_reflections = flex.reflection_table()
        for expt, refl in zip(
            self.experiments, self.reflections.split_by_experiment_id()
        ):
            # extract experiment details
            detector = expt.detector
            panels = list(detector)
            panel_size_px = [p.get_image_size() for p in panels]
            pixel_size_mm = [p.get_pixel_size()[0] for p in panels]
            detector_dist_mm = [p.get_distance() for p in panels]
            beam = expt.beam
            wavelength_ang = beam.get_wavelength()

            # exclude reflections with no foreground pixels
            refl_valid = refl.select(
                refl["num_pixels.valid"] > 0 and refl["num_pixels.foreground"] > 0
            )
            refl_zero = refl_valid.select(refl_valid["intensity.sum.value"] == 0)
            refl_nonzero = refl_valid.select(refl_valid["intensity.sum.value"] != 0)

            def correct(refl_sele, smart_sigmas=True):
                kapton_correction = image_kapton_correction(
                    panel_size_px=panel_size_px,
                    pixel_size_mm=pixel_size_mm,
                    detector_dist_mm=detector_dist_mm,
                    wavelength_ang=wavelength_ang,
                    reflections_sele=refl_sele,
                    params=self.params,
                    expt=expt,
                    refl=refl,
                    smart_sigmas=smart_sigmas,
                    logger=self.logger,
                )

                k_corr, k_sigmas = kapton_correction()
                refl_sele["kapton_absorption_correction"] = k_corr
                if smart_sigmas:
                    refl_sele["kapton_absorption_correction_sigmas"] = k_sigmas
                    # apply corrections and propagate error
                    # term1 = (sig(C)/C)^2
                    # term2 = (sig(Imeas)/Imeas)^2
                    # I' = C*I
                    # sig^2(I') = (I')^2*(term1 + term2)
                    integrated_data = refl_sele["intensity.sum.value"]
                    integrated_variance = refl_sele["intensity.sum.variance"]
                    integrated_sigma = flex.sqrt(integrated_variance)
                    term1 = flex.pow(k_sigmas / k_corr, 2)
                    term2 = flex.pow(integrated_sigma / integrated_data, 2)
                    integrated_data *= k_corr
                    integrated_variance = flex.pow(integrated_data, 2) * (term1 + term2)
                    refl_sele["intensity.sum.value"] = integrated_data
                    refl_sele["intensity.sum.variance"] = integrated_variance
                    # order is purposeful: the two lines above require that integrated_data
                    # has already been corrected!
                else:
                    refl_sele["intensity.sum.value"] *= k_corr
                    refl_sele["intensity.sum.variance"] *= flex.pow2(k_corr)
                return refl_sele

            if len(refl_zero) > 0 and self.params.smart_sigmas:
                # process nonzero intensity reflections with smart sigmas as requested
                # but turn them off for zero intensity reflections to avoid a division by zero
                # during error propagation. Not at all certain this is the best way.
                self.corrected_reflections.extend(
                    correct(refl_nonzero, smart_sigmas=True)
                )
                self.corrected_reflections.extend(
                    correct(refl_zero, smart_sigmas=False)
                )
            else:
                self.corrected_reflections.extend(
                    correct(refl_valid, smart_sigmas=self.params.smart_sigmas)
                )

        return self.experiments, self.corrected_reflections



"""Target function for cosym analysis."""

from __future__ import annotations

import copy
import itertools
import logging

import numpy as np
import pandas as pd
from orderedset import OrderedSet

import cctbx.sgtbx.cosets
from cctbx import miller, sgtbx
from cctbx.array_family import flex

logger = logging.getLogger(__name__)


class Target:
    """Target function for cosym analysis.

    Attributes:
      dim (int): The number of dimensions used in the analysis.
    """

    def __init__(
        self,
        intensities,
        lattice_ids,
        weights=None,
        min_pairs=3,
        lattice_group=None,
        dimensions=None,
    ):
        r"""Initialise a Target object.

        Args:
          intensities (cctbx.miller.array): The intensities on which to perform
            cosym analysis.
          lattice_ids (np.ndarray): An array of equal size to
            `intensities` which maps each reflection to a given lattice (dataset).
          weights (str): Optionally include weights in the target function.
            Allowed values are `None`, "count" and "standard_error". The default
            is to use no weights. If "count" is set, then weights are equal to the
            number of pairs of reflections used in calculating each value of the
            rij matrix. If "standard_error" is used, then weights are defined as
            :math:`w_{ij} = 1/s`, where :math:`s = \sqrt{(1-r_{ij}^2)/(n-2)}`.
            See also http://www.sjsu.edu/faculty/gerstman/StatPrimer/correlation.pdf.
          min_pairs (int): Only calculate the correlation coefficient between two
            datasets if they have more than `min_pairs` of common reflections.
          lattice_group (cctbx.sgtbx.space_group): Optionally set the lattice
            group to be used in the analysis.
          dimensions (int): Optionally override the number of dimensions to be used
            in the analysis. If not set, then the number of dimensions used is
            equal to the greater of 2 or the number of symmetry operations in the
            lattice group.
        """
        if weights is not None:
            assert weights in ("count", "standard_error")
        self._weights = weights
        self._min_pairs = min_pairs

        data = intensities.customized_copy(anomalous_flag=False)
        cb_op_to_primitive = data.change_of_basis_op_to_primitive_setting()
        data = data.change_basis(cb_op_to_primitive).map_to_asu()

        # Convert to uint64 avoids crashes on Windows when later constructing
        # flex.size_t (https://github.com/cctbx/cctbx_project/issues/591)
        order = lattice_ids.argsort().astype(np.uint64)
        sorted_data = data.data().select(flex.size_t(order))
        sorted_indices = data.indices().select(flex.size_t(order))
        self._lattice_ids = lattice_ids[order]
        self._data = data.customized_copy(indices=sorted_indices, data=sorted_data)
        assert isinstance(self._data.indices(), type(flex.miller_index()))
        assert isinstance(self._data.data(), type(flex.double()))

        # construct a lookup for the separate lattices
        self._lattices = np.array(
            [
                np.where(self._lattice_ids == i)[0][0]
                for i in np.unique(self._lattice_ids)
            ]
        )

        self.sym_ops = OrderedSet(["x,y,z"])
        self._lattice_group = lattice_group
        self.sym_ops.update(op.as_xyz() for op in self._generate_twin_operators())
        if dimensions is None:
            dimensions = max(2, len(self.sym_ops))
        self.set_dimensions(dimensions)

        self._lattice_group = copy.deepcopy(self._data.space_group())
        for sym_op in self.sym_ops:
            self._lattice_group.expand_smx(sym_op)
        self._patterson_group = self._lattice_group.build_derived_patterson_group()

        logger.debug(
            "Lattice group: %s (%i symops)",
            self._lattice_group.info().symbol_and_number(),
            len(self._lattice_group),
        )
        logger.debug(
            "Patterson group: %s", self._patterson_group.info().symbol_and_number()
        )

        self.rij_matrix, self.wij_matrix = self._compute_rij_wij()

    def set_dimensions(self, dimensions):
        """Set the number of dimensions for analysis.

        Args:
          dimensions (int): The number of dimensions to be used.
        """
        self.dim = dimensions

    def _generate_twin_operators(self, lattice_symmetry_max_delta=5.0):
        # see also mmtbx.scaling.twin_analyses.twin_laws
        if self._lattice_group is None:
            cb_op_to_minimum_cell = self._data.change_of_basis_op_to_minimum_cell()
            minimum_cell_symmetry = self._data.crystal_symmetry().change_basis(
                cb_op=cb_op_to_minimum_cell
            )
            self._lattice_group = sgtbx.lattice_symmetry.group(
                reduced_cell=minimum_cell_symmetry.unit_cell(),
                max_delta=lattice_symmetry_max_delta,
            )
            intensity_symmetry = minimum_cell_symmetry.reflection_intensity_symmetry(
                anomalous_flag=self._data.anomalous_flag()
            )
            cb_op = cb_op_to_minimum_cell.inverse()
        else:
            cb_op = sgtbx.change_of_basis_op()
            intensity_symmetry = self._data.reflection_intensity_symmetry()

        operators = []
        for partition in cctbx.sgtbx.cosets.left_decomposition(
            g=self._lattice_group,
            h=intensity_symmetry.space_group()
            .build_derived_acentric_group()
            .make_tidy(),
        ).partitions[1:]:
            if partition[0].r().determinant() > 0:
                operators.append(cb_op.apply(partition[0]))

        return operators

    def _compute_rij_wij(self, use_cache=True):
        """Compute the rij_wij matrix.

        Rij is a symmetric matrix of size (n x m, n x m), where n is the number of
        datasets and m is the number of symmetry operations.

        It is composed of (m, m) blocks of size (n, n), where each block contains the
        correlation coefficients between cb_op_k applied to datasets 1..N with
        cb_op_kk applied to datasets 1.. N.

        If `use_cache=True`, then an optimisation is made to reflect the fact some elements
        of the matrix are equivalent, i.e.:
            CC[(a, cb_op_k), (b, cb_op_kk)] == CC[(a,), (b, cb_op_k.inverse() * cb_op_kk)]

        """
        n_lattices = len(self._lattices)
        n_sym_ops = len(self.sym_ops)

        # Pre-calculate miller indices after application of each cb_op. Only calculate
        # this once per cb_op instead of on-the-fly every time we need it.
        indices = {}
        epsilons = {}
        space_group_type = self._data.space_group().type()
        for cb_op in self.sym_ops:
            cb_op = sgtbx.change_of_basis_op(cb_op)
            indices_reindexed = cb_op.apply(self._data.indices())
            miller.map_to_asu(space_group_type, False, indices_reindexed)
            cb_op_str = cb_op.as_xyz()
            indices[cb_op_str] = np.array(
                [
                    h.iround().as_numpy_array()
                    for h in indices_reindexed.as_vec3_double().parts()
                ]
            ).transpose()
            epsilons[cb_op_str] = self._patterson_group.epsilon(
                indices_reindexed
            ).as_numpy_array()
        intensities = self._data.data().as_numpy_array()

        # Map indices to an array of flat 1d indices which can later be used for
        # matching pairs of indices
        offset = -np.min(np.concatenate(list(indices.values())), axis=0)
        dims = np.max(np.concatenate(list(indices.values())), axis=0) + offset + 1
        for cb_op, hkl in indices.items():
            indices[cb_op] = np.ravel_multi_index((hkl + offset).T, dims)

        # Create an empty 2D array of shape (m * n, L), where m is the number of sym
        # ops, n is the number of lattices, and L is the number of unique miller indices
        all_intensities = np.empty((n_sym_ops * n_lattices, np.prod(dims)))

        # Populate all_intensities with intensity values, filling absent intensities
        # with np.nan
        all_intensities.fill(np.nan)
        slices = np.append(self._lattices, intensities.size)
        slices = list(map(slice, slices[:-1], slices[1:]))
        for i, (mil_ind, eps) in enumerate(zip(indices.values(), epsilons.values())):
            for j, selection in enumerate(slices):
                # map (i, j) to a column in all_intensities
                column = np.ravel_multi_index((i, j), (n_sym_ops, n_lattices))
                epsilon_equals_one = eps[selection] == 1
                valid_mil_ind = mil_ind[selection][epsilon_equals_one]
                valid_intensities = intensities[selection][epsilon_equals_one]
                all_intensities[column, valid_mil_ind] = valid_intensities

        # Ideally we would use `np.ma.corrcoef` here, but it is broken, so use
        # pd.DataFrame.corr() instead (see numpy/numpy#15601)
        rij = (
            pd.DataFrame(all_intensities)
            .T.dropna(how="all")
            .corr(min_periods=self._min_pairs)
            .values
        )
        # Set any NaN correlation coefficients to zero
        np.nan_to_num(rij, copy=False)
        # Cosym does not make use of the on-diagonal correlation coefficients
        np.fill_diagonal(rij, 0)

        if self._weights:
            wij = np.zeros_like(rij)
            right_up = np.triu_indices_from(wij, k=1)

            # For each correlation coefficient, set the weight equal to the size of
            # the sample used to calculate that coefficient
            pairwise_combos = itertools.combinations(np.isfinite(all_intensities), 2)
            sample_size = lambda x, y: np.count_nonzero(x & y)
            wij[right_up] = list(itertools.starmap(sample_size, pairwise_combos))

            if self._weights == "standard_error":
                # Set each weights as the reciprocal of the standard error on the
                # corresponding correlation coefficient
                # http://www.sjsu.edu/faculty/gerstman/StatPrimer/correlation.pdf
                with np.errstate(divide="ignore", invalid="ignore"):
                    reciprocal_se = np.sqrt(
                        (wij[right_up] - 2) / (1 - np.square(rij[right_up]))
                    )

                wij[right_up] = np.where(wij[right_up] > 2, reciprocal_se, 0)

            # Symmetrise the wij matrix
            wij += wij.T
        else:
            wij = None

        return rij, wij

    def compute_functional(self, x: np.ndarray) -> float:
        """Compute the target function at coordinates `x`.

        Args:
          x (np.ndarray):
            a flattened list of the N-dimensional vectors, i.e. coordinates in
            the first dimension are stored first, followed by the coordinates in
            the second dimension, etc.

        Returns:
          f (float): The value of the target function at coordinates `x`.
        """
        assert (x.size // self.dim) == (len(self._lattices) * len(self.sym_ops))
        x = x.reshape((self.dim, x.size // self.dim))
        elements = np.square(self.rij_matrix - x.T @ x)
        if self.wij_matrix is not None:
            np.multiply(self.wij_matrix, elements, out=elements)
        f = 0.5 * elements.sum()
        return f

    def compute_gradients_fd(self, x: np.ndarray, eps=1e-6) -> np.ndarray:
        """Compute the gradients at coordinates `x` using finite differences.

        Args:
          x (np.ndarray):
            a flattened list of the N-dimensional vectors, i.e. coordinates in
            the first dimension are stored first, followed by the coordinates in
            the second dimension, etc.
          eps (float):
            The value of epsilon to use in finite difference calculations.

        Returns:
          grad (np.ndarray):
          The gradients of the target function with respect to the parameters.
        """
        x = copy.deepcopy(x)
        grad = np.zeros(x.shape)
        for i in range(x.size):
            x[i] += eps  # x + eps
            fp = self.compute_functional(x)
            x[i] -= 2 * eps  # x - eps
            fm = self.compute_functional(x)
            x[i] += eps  # reset to original values
            grad[i] += (fp - fm) / (2 * eps)
        return grad

    def compute_gradients(self, x: np.ndarray) -> np.ndarray:
        """Compute the gradients of the target function at coordinates `x`.

        Args:
          x (np.ndarray):
            a flattened list of the N-dimensional vectors, i.e. coordinates in
            the first dimension are stored first, followed by the coordinates in
            the second dimension, etc.

        Returns:
          Tuple[float, np.ndarray]:
          f: The value of the target function at coordinates `x`.
          grad: The gradients of the target function with respect to the parameters.
        """
        x = x.reshape((self.dim, x.size // self.dim))
        if self.wij_matrix is not None:
            wrij_matrix = np.multiply(self.wij_matrix, self.rij_matrix)
            grad = -2 * x @ (wrij_matrix - np.multiply(self.wij_matrix, x.T @ x))
        else:
            grad = -2 * x @ (self.rij_matrix - x.T @ x)
        return grad.flatten()

    def curvatures(self, x: np.ndarray) -> np.ndarray:
        """Compute the curvature of the target function at coordinates `x`.

        Args:
          x (np.ndarray):
            a flattened list of the N-dimensional vectors, i.e. coordinates in
            the first dimension are stored first, followed by the coordinates in
            the second dimension, etc.

        Returns:
          curvs (np.ndarray):
          The curvature of the target function with respect to the parameters.
        """
        if self.wij_matrix is not None:
            wij = self.wij_matrix
        else:
            wij = np.ones(self.rij_matrix.shape)
        x = x.reshape((self.dim, x.size // self.dim))
        curvs = 2 * np.square(x) @ wij
        return curvs.flatten()

    def curvatures_fd(self, x: np.ndarray, eps=1e-6) -> np.ndarray:
        """Compute the curvatures at coordinates `x` using finite differences.

        Args:
          x (np.ndarray):
            a flattened list of the N-dimensional vectors, i.e. coordinates in
            the first dimension are stored first, followed by the coordinates in
            the second dimension, etc.
          eps (float):
            The value of epsilon to use in finite difference calculations.

        Returns:
          curvs (np.ndarray):
          The curvature of the target function with respect to the parameters.
        """
        x = copy.deepcopy(x)
        f = self.compute_functional(x)
        curvs = np.zeros(x.shape)
        for i in range(x.size):
            x[i] += eps  # x + eps
            fp = self.compute_functional(x)
            x[i] -= 2 * eps  # x - eps
            fm = self.compute_functional(x)
            x[i] += eps  # reset to original values
            curvs[i] += (fm - 2 * f + fp) / (eps**2)
        return curvs



from __future__ import annotations

import math

import wx


class RingSettingsFrame(wx.MiniFrame):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        szr = wx.BoxSizer(wx.VERTICAL)
        panel = RingSettingsPanel(self)
        self.SetSizer(szr)
        szr.Add(panel, 1, wx.EXPAND)
        szr.Fit(panel)
        self.panel = panel
        self.sizer = szr
        self.Fit()
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)


class RingSettingsPanel(wx.Panel):
    def __init__(self, *args, **kwds):
        # XXX Support several rings.  Plot radial distribution somewhere
        # (not here), but maybe distribution along ring.  Drop-down menu
        # for ring center, and button to reset to beam center.

        super().__init__(*args, **kwds)

        # Needed to draw and delete the rings.  XXX Applies to
        # calibration_frame as well?
        self._pyslip = self.GetParent().GetParent().pyslip

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        # Number of decimal digits for distances.
        self.digits = 2

        # Distance control XXX probably does not belong here
        #    print "DISTANCE",self.GetParent().GetParent().viewer._img
        #    box = wx.BoxSizer(wx.HORIZONTAL)
        #    from wxtbx.phil_controls.floatctrl import FloatCtrl
        #    from wxtbx.phil_controls import EVT_PHIL_CONTROL
        #    self.distance_ctrl = FloatCtrl(self, pos=(300,180), size=(80,-1),
        #    value=80.00,
        #    name="Detector Distance")
        #    self.distance_ctrl.SetMax(1000)
        #    self.distance_ctrl.SetMin(5)
        #    self.distance_ctrl.SetOptional(False)
        #    box.Add(self.distance_ctrl, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5)
        #    txtd = wx.StaticText(self, label="Detector Distance")
        #    box.Add(txtd, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, 5)
        #    s.Add(box)

        from wx.lib.agw.floatspin import EVT_FLOATSPIN, FloatSpin

        # XXX Should really make value be in Aangstroem resolution, and
        # have a non-linear slider.
        self._radius = 100
        self._center = [0, 0]
        radius_max = 2000
        radius_min = 10

        # Radius controls.
        box = wx.BoxSizer(wx.HORIZONTAL)

        self.slider = wx.Slider(
            self,
            maxValue=radius_max,
            minValue=radius_min,
            size=(250, -1),
            style=wx.SL_AUTOTICKS | wx.SL_HORIZONTAL,
            value=self._radius,
        )
        box.Add(
            self.slider, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )
        self.Bind(wx.EVT_SLIDER, self.OnSlide, self.slider)

        self.spinner = FloatSpin(
            self,
            digits=self.digits,
            max_val=radius_max,
            min_val=radius_min,
            value=self._radius,
        )
        box.Add(
            self.spinner, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpin, self.spinner)

        self.auto = wx.Button(self, label="Auto fit")
        self.Bind(wx.EVT_BUTTON, self.OnAutoFit, self.auto)
        box.Add(
            self.auto, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )

        sizer.Add(box)

        # Centering controls.
        box = wx.BoxSizer(wx.HORIZONTAL)

        self.spinner_fast = FloatSpin(
            self, digits=self.digits, name="fast_ctrl", value=self._center[0]
        )
        box.Add(
            self.spinner_fast,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="Center fast"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinCenter, self.spinner_fast)

        self.spinner_slow = FloatSpin(
            self, digits=self.digits, name="slow_ctrl", value=self._center[1]
        )
        box.Add(
            self.spinner_slow,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="Center slow"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinCenter, self.spinner_slow)

        self.clear_button = wx.Button(self, -1, "Clear")
        box.Add(self.clear_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_BUTTON, self.OnClear, self.clear_button)

        sizer.Add(box)

        self.DrawRing()

    def __del__(self):
        if hasattr(self, "_ring_layer") and self._ring_layer is not None:
            self._pyslip.DeleteLayer(self._ring_layer)
            self._ring_layer = None

    def OnSlide(self, event):
        # Keep slider and spinner synchronized.
        obj = event.EventObject  # XXX Is this construct sane?  See below
        # and in calibration_frame, too!
        self._radius = obj.GetValue()
        self.spinner.SetValue(self._radius)

        self.DrawRing()

    def OnAutoFit(self, event):
        jitter = 6

        detector = self._pyslip.tiles.raw_image.get_detector()
        beam = self._pyslip.tiles.raw_image.get_beam()
        # FIXME assumes all detector elements use the same millimeter-to-pixel convention
        if detector[0].get_distance() > 0:
            h = detector.hierarchy()
            if len(h) > 0:
                beam_pixel_fast, beam_pixel_slow = detector[0].millimeter_to_pixel(
                    detector.hierarchy().get_beam_centre(beam.get_s0())
                )
            else:
                beam_pixel_fast, beam_pixel_slow = detector[0].millimeter_to_pixel(
                    detector[0].get_beam_centre(beam.get_s0())
                )

        avg_distance = -sum(p.get_distance() for p in detector) / len(detector)

        beam_pixel_fast += self._center[0]
        beam_pixel_slow += self._center[1]

        def PointsOnCircle(center, radius, count):
            for r in range(count):
                t = (r / count) * 2 * math.pi
                yield (
                    center[0] + (radius * math.cos(t)),
                    center[1] + (radius * math.sin(t)),
                )

        best = float("-inf")
        bestc = [self._center[0], self._center[1]]
        bestr = self._radius

        image_data = self._pyslip.tiles.raw_image.get_image_data()
        if not isinstance(image_data, tuple):
            image_data = (image_data,)

        for j in range(-jitter, jitter, 1):
            j /= 2
            for i in range(-jitter, jitter, 1):
                i /= 2
                for r in range(-jitter, jitter, 1):
                    r /= 2
                    total = 0.0
                    for point in PointsOnCircle(
                        (beam_pixel_fast + i, beam_pixel_slow + j),
                        self._radius + r,
                        360,
                    ):
                        mm = detector[0].pixel_to_millimeter(point)
                        mm = (mm[0], mm[1], avg_distance)
                        pid = detector.get_panel_intersection(mm)
                        if pid >= 0:
                            px = detector[pid].get_ray_intersection_px(mm)
                            px = [int(round(px[0])), int(round(px[1]))]
                            data = image_data[pid]
                            if (
                                px[0] >= 0
                                and px[0] < data.focus()[1]
                                and px[1] >= 0
                                and px[1] < data.focus()[0]
                            ):
                                total += data[px[1], px[0]]
                    if total > best:
                        best = total
                        bestc = [self._center[0] + i, self._center[1] + j]
                        bestr = self._radius + r
                    print(f"r: {r: 3.1f}, i: {i: 3.1f}, j: {j: 3.1f}, best: {best:f}")
        print("DONE", bestc, bestr)
        self._radius = bestr
        self._center = bestc

        self.spinner.SetValue(bestr)
        self.spinner_fast.SetValue(bestc[0])
        self.spinner_slow.SetValue(bestc[1])

        self.DrawRing()

    def OnSpin(self, event):
        # Keep slider and spinner synchronized.  XXX OnSpinRadius()?
        obj = event.EventObject
        self._radius = obj.GetValue()
        self.slider.SetValue(self._radius)

        self.DrawRing()

    def OnSpinCenter(self, event):
        obj = event.EventObject
        name = obj.GetName()

        if name == "fast_ctrl":
            self._center[0] = obj.GetValue()
        elif name == "slow_ctrl":
            self._center[1] = obj.GetValue()

        self.DrawRing()

    def OnClear(self, event):
        self.__del__()

    def _draw_ring_layer(self, dc, data, map_rel):
        """Draw a points layer.

        dc       the device context to draw on
        data     an iterable of point tuples:
                 (x, y, place, radius, colour, x_off, y_off, pdata)
        map_rel  points relative to map if True, MUST BE TRUE for lightweight
        Assumes all points are the same colour, saving 100's of ms.
        """

        assert map_rel is True
        if len(data) == 0:
            return
        (lon, lat, place, radius, colour, x_off, y_off, pdata) = data[0]

        scale = 2**self._pyslip.tiles.zoom_level

        # Draw points on map/view, using transparency if implemented.
        try:
            dc = wx.GCDC(dc)
        except NotImplementedError:
            pass
        dc.SetPen(wx.Pen(colour))
        dc.SetBrush(wx.Brush(colour, wx.TRANSPARENT))
        for (lon, lat, place, radius, colour, x_off, y_off, pdata) in data:
            (x, y) = self._pyslip.ConvertGeo2View((lon, lat))
            dc.DrawCircle(x, y, radius * scale)

    def DrawRing(self):

        xrayframe = self.GetParent().GetParent()
        panel_id, beam_pixel_fast, beam_pixel_slow = xrayframe.get_beam_center_px()

        (
            beam_pixel_slow,
            beam_pixel_fast,
        ) = xrayframe.pyslip.tiles.flex_image.tile_readout_to_picture(
            panel_id, beam_pixel_slow - 0.5, beam_pixel_fast - 0.5
        )

        center = self._pyslip.tiles.picture_fast_slow_to_map_relative(
            beam_pixel_fast + self._center[0], beam_pixel_slow + self._center[1]
        )

        # XXX Transparency?
        ring_data = [(center[0], center[1], {"colour": "red", "radius": self._radius})]

        # Remove the old ring layer, and draw a new one.  XXX Why
        # disappears at highest levels?
        if hasattr(self, "_ring_layer") and self._ring_layer is not None:
            self._pyslip.DeleteLayer(self._ring_layer)
            self._ring_layer = None
        self._ring_layer = self._pyslip.AddPointLayer(
            ring_data,
            map_rel=True,
            visible=True,
            show_levels=[-3, -2, -1, 0, 1, 2, 3, 4, 5],
            renderer=self._draw_ring_layer,
            name="<ring_layer>",
        )
